from typing import Any, Dict, Sequence, Type, Union, cast

import copy
import glob
import importlib
import inspect
import itertools
import os
import pickle
import re
import tempfile

import joblib
import numpy as np
import numpy.ma as npma
import pandas as pd
import sklearn
import sklearn.datasets
import sklearn.dummy
import sklearn.model_selection as model_selection
from ConfigSpace.configuration_space import Configuration
from dask.distributed import Client
from joblib import cpu_count
from sklearn.base import ClassifierMixin, RegressorMixin, clone, is_classifier
from smac.tae import StatusType

import autosklearn.estimators  # noqa F401
import autosklearn.pipeline.util as putil
from autosklearn.automl import AutoMLClassifier
from autosklearn.data.validation import InputValidator
from autosklearn.ensemble_building.run import Run
from autosklearn.ensembles import EnsembleSelection, MultiObjectiveDummyEnsemble
from autosklearn.estimators import (
    AutoSklearnClassifier,
    AutoSklearnEstimator,
    AutoSklearnRegressor,
)
from autosklearn.experimental.askl2 import AutoSklearn2Classifier
from autosklearn.metrics import accuracy, f1_macro, mean_squared_error, r2
from autosklearn.smbo import get_smac_object

import pytest
import unittest
import unittest.mock

import test.conftest
from test.test_automl.automl_utils import (
    count_succeses,
    include_single_scores,
    includes_all_scores,
    includes_train_scores,
    performance_over_time_is_plausible,
    print_debug_information,
)


def test_fit_n_jobs(tmp_dir):
    X_train, Y_train, X_test, Y_test = putil.get_dataset("breast_cancer")

    # test parallel Classifier to predict classes, not only indices
    Y_train += 1
    Y_test += 1

    n_jobs = 2

    class get_smac_object_wrapper:
        def __call__(self, *args, **kwargs):
            self.n_jobs = kwargs["n_jobs"]
            smac = get_smac_object(*args, **kwargs)
            self.dask_n_jobs = smac.solver.tae_runner.n_workers
            self.dask_client_n_jobs = len(
                smac.solver.tae_runner.client.scheduler_info()["workers"]
            )
            return smac

    get_smac_object_wrapper_instance = get_smac_object_wrapper()

    automl = AutoSklearnClassifier(
        delete_tmp_folder_after_terminate=False,
        time_left_for_this_task=30,
        per_run_time_limit=5,
        tmp_folder=os.path.join(tmp_dir, "backend"),
        seed=1,
        n_jobs=n_jobs,
        get_smac_object_callback=get_smac_object_wrapper_instance,
        max_models_on_disc=None,
    )
    automl.fit(X_train, Y_train)

    # Test that the argument is correctly passed to SMAC
    assert get_smac_object_wrapper_instance.n_jobs == n_jobs
    assert get_smac_object_wrapper_instance.dask_n_jobs == n_jobs
    assert get_smac_object_wrapper_instance.dask_client_n_jobs == n_jobs

    runhistory_data = automl.automl_.runhistory_.data

    successful_runs = {
        run_value.additional_info["num_run"]
        for run_value in runhistory_data.values()
        if (
            run_value.additional_info is not None
            and "num_run" in run_value.additional_info
            and run_value.status == StatusType.SUCCESS
        )
    }

    available_predictions = set()
    predictions = glob.glob(
        os.path.join(
            automl.automl_._backend.get_runs_directory(), "*", "predictions_ensemble_*"
        )
    )
    seeds = set()
    for prediction in predictions:
        prediction = os.path.split(prediction)[1]
        match = re.match(Run.RE_MODEL_PREDICTION_FILE, prediction)
        if match:
            print(match)
            seed, num_run, _ = match.groups()
            available_predictions.add(int(num_run))
            seeds.add(int(seed))

    # Remove the dummy prediction, it is not part of the runhistory
    if 1 in available_predictions:
        available_predictions.remove(1)

    # Make sure all predictions available are associated with a successful run
    # Don't want a rogue prediction file
    assert available_predictions <= successful_runs

    assert len(seeds) == 1

    ensemble_dir = automl.automl_._backend.get_ensemble_dir()
    ensembles = os.listdir(ensemble_dir)

    seeds = set()
    for ensemble_file in ensembles:
        seeds.add(int(ensemble_file.split(".")[0].split("_")[0]))
    assert len(seeds) == 1

    assert count_succeses(automl.cv_results_) > 0
    assert includes_train_scores(automl.performance_over_time_.columns) is True
    assert performance_over_time_is_plausible(automl.performance_over_time_) is True


def test_feat_type_wrong_arguments():

    # Every Auto-Sklearn estimator has a backend, that allows a single
    # call to fit
    X = np.zeros((100, 100))
    y = np.zeros((100,))

    cls = AutoSklearnClassifier()
    expected_msg = r".*feat_type does not have same number of "
    "variables as X has features. 1 vs 100.*"
    with pytest.raises(ValueError, match=expected_msg):
        cls.fit(X=X, y=y, feat_type=[True])

    cls = AutoSklearnClassifier()
    expected_msg = r".*feat_type must only contain strings.*"
    with pytest.raises(ValueError, match=expected_msg):
        cls.fit(X=X, y=y, feat_type=[True] * 100)

    cls = AutoSklearnClassifier()
    expected_msg = r".*Only `Categorical`, `Numerical` and `String` are"
    "valid feature types, you passed `Car`.*"
    with pytest.raises(ValueError, match=expected_msg):
        cls.fit(X=X, y=y, feat_type=["Car"] * 100)


# Mock AutoSklearnEstimator.fit so the test doesn't actually run fit().
@unittest.mock.patch("autosklearn.estimators.AutoSklearnEstimator.fit")
def test_type_of_target(mock_estimator):
    # Test that classifier raises error for illegal target types.
    X = np.array(
        [
            [1, 2],
            [2, 3],
            [3, 4],
            [4, 5],
        ]
    )
    # Possible target types
    y_binary = np.array([0, 0, 1, 1])
    y_continuous = np.array([0.1, 1.3, 2.1, 4.0])
    y_multiclass = np.array([0, 1, 2, 0])
    y_multilabel = np.array(
        [
            [0, 1],
            [1, 1],
            [1, 0],
            [0, 0],
        ]
    )
    y_multiclass_multioutput = np.array(
        [
            [0, 1],
            [1, 3],
            [2, 2],
            [5, 3],
        ]
    )
    y_continuous_multioutput = np.array(
        [
            [0.1, 1.5],
            [1.2, 3.5],
            [2.7, 2.7],
            [5.5, 3.9],
        ]
    )

    cls = AutoSklearnClassifier()
    cls.automl_ = unittest.mock.Mock()
    cls.automl_.InputValidator = unittest.mock.Mock()
    cls.automl_.InputValidator.target_validator = unittest.mock.Mock()

    # Illegal target types for classification: continuous,
    # multiclass-multioutput, continuous-multioutput.
    expected_msg = r".*Classification with data of type"
    " multiclass-multioutput is not supported.*"
    with pytest.raises(ValueError, match=expected_msg):
        cls.fit(X=X, y=y_multiclass_multioutput)

    expected_msg = r".*Classification with data of type"
    " continuous is not supported.*"
    with pytest.raises(ValueError, match=expected_msg):
        cls.fit(X=X, y=y_continuous)

    expected_msg = r".*Classification with data of type"
    " continuous-multioutput is not supported.*"
    with pytest.raises(ValueError, match=expected_msg):
        cls.fit(X=X, y=y_continuous_multioutput)

    # Legal target types for classification: binary, multiclass,
    # multilabel-indicator.
    try:
        cls.fit(X, y_binary)
    except ValueError:
        pytest.fail("cls.fit() raised ValueError while fitting " "binary targets")

    try:
        cls.fit(X, y_multiclass)
    except ValueError:
        pytest.fail("cls.fit() raised ValueError while fitting " "multiclass targets")

    try:
        cls.fit(X, y_multilabel)
    except ValueError:
        pytest.fail(
            "cls.fit() raised ValueError while fitting " "multilabel-indicator targets"
        )

    # Test that regressor raises error for illegal target types.
    reg = AutoSklearnRegressor()
    # Illegal target types for regression: multilabel-indicator
    # multiclass-multioutput
    expected_msg = r".*Regression with data of type"
    " multilabel-indicator is not supported.*"
    with pytest.raises(ValueError, match=expected_msg):
        reg.fit(
            X=X,
            y=y_multilabel,
        )

    expected_msg = r".*Regression with data of type"
    " multiclass-multioutput is not supported.*"
    with pytest.raises(ValueError, match=expected_msg):
        reg.fit(
            X=X,
            y=y_multiclass_multioutput,
        )

    # Legal target types: continuous, multiclass,
    # continuous-multioutput,
    # binary
    try:
        reg.fit(X, y_continuous)
    except ValueError:
        pytest.fail("reg.fit() raised ValueError while fitting " "continuous targets")

    try:
        reg.fit(X, y_multiclass)
    except ValueError:
        pytest.fail("reg.fit() raised ValueError while fitting " "multiclass targets")

    try:
        reg.fit(X, y_continuous_multioutput)
    except ValueError:
        pytest.fail(
            "reg.fit() raised ValueError while fitting "
            "continuous_multioutput targets"
        )

    try:
        reg.fit(X, y_binary)
    except ValueError:
        pytest.fail("reg.fit() raised ValueError while fitting " "binary targets")


def test_performance_over_time_no_ensemble(tmp_dir):
    X_train, Y_train, X_test, Y_test = putil.get_dataset("iris")

    cls = AutoSklearnClassifier(
        time_left_for_this_task=30,
        per_run_time_limit=5,
        tmp_folder=os.path.join(tmp_dir, "backend"),
        seed=1,
        ensemble_class=None,
        initial_configurations_via_metalearning=0,
    )

    cls.fit(X_train, Y_train, X_test, Y_test)

    performance_over_time = cls.performance_over_time_
    assert include_single_scores(performance_over_time.columns) is True
    assert performance_over_time_is_plausible(performance_over_time) is True


def test_cv_results(tmp_dir):
    # TODO restructure and actually use real SMAC output from a long run
    # to do this unittest!
    X_train, Y_train, X_test, Y_test = putil.get_dataset("iris")

    cls = AutoSklearnClassifier(
        time_left_for_this_task=30,
        per_run_time_limit=5,
        tmp_folder=os.path.join(tmp_dir, "backend"),
        seed=1,
        initial_configurations_via_metalearning=0,
        scoring_functions=[autosklearn.metrics.precision, autosklearn.metrics.roc_auc],
    )

    params = cls.get_params()
    original_params = copy.deepcopy(params)

    cls.fit(X_train, Y_train)

    cv_results = cls.cv_results_
    assert isinstance(cv_results, dict), type(cv_results)
    assert isinstance(cv_results["mean_test_score"], np.ndarray), type(
        cv_results["mean_test_score"]
    )
    assert isinstance(cv_results["mean_fit_time"], np.ndarray), type(
        cv_results["mean_fit_time"]
    )
    assert isinstance(cv_results["params"], list), type(cv_results["params"])
    assert isinstance(cv_results["rank_test_scores"], np.ndarray), type(
        cv_results["rank_test_scores"]
    )
    assert isinstance(cv_results["metric_precision"], npma.MaskedArray), type(
        cv_results["metric_precision"]
    )
    assert isinstance(cv_results["metric_roc_auc"], npma.MaskedArray), type(
        cv_results["metric_roc_auc"]
    )
    cv_result_items = [
        isinstance(val, npma.MaskedArray)
        for key, val in cv_results.items()
        if key.startswith("param_")
    ]
    assert all(cv_result_items), cv_results.items()

    # Compare the state of the model parameters with the original parameters
    new_params = clone(cls).get_params()
    for param_name, original_value in original_params.items():
        new_value = new_params[param_name]

        # Taken from Sklearn code:
        # We should never change or mutate the internal state of input
        # parameters by default. To check this we use the joblib.hash function
        # that introspects recursively any subobjects to compute a checksum.
        # The only exception to this rule of immutable constructor parameters
        # is possible RandomState instance but in this check we explicitly
        # fixed the random_state params recursively to be integer seeds.
        assert joblib.hash(new_value) == joblib.hash(original_value), (
            "Estimator %s should not change or mutate "
            " the parameter %s from %s to %s during fit."
            % (cls, param_name, original_value, new_value)
        )

    # Comply with https://scikit-learn.org/dev/glossary.html#term-classes
    is_classifier(cls)
    assert hasattr(cls, "classes_")


def test_cv_results_multi_objective(tmp_dir):
    # TODO restructure and actually use real SMAC output from a long run
    # to do this unittest!
    X_train, Y_train, X_test, Y_test = putil.get_dataset("iris")

    cls = AutoSklearnClassifier(
        time_left_for_this_task=30,
        per_run_time_limit=5,
        tmp_folder=os.path.join(tmp_dir, "backend"),
        seed=1,
        initial_configurations_via_metalearning=0,
        metric=[autosklearn.metrics.precision_macro, autosklearn.metrics.recall_macro],
        scoring_functions=[
            autosklearn.metrics.accuracy,
            autosklearn.metrics.balanced_accuracy,
        ],
    )

    params = cls.get_params()
    original_params = copy.deepcopy(params)

    cls.fit(X_train, Y_train)

    cv_results = cls.cv_results_
    assert isinstance(cv_results, dict), type(cv_results)
    assert "mean_test_score" not in cv_results
    assert "rank_test_scores" not in cv_results
    for expected_column in (
        "mean_test_precision_macro",
        "mean_test_recall_macro",
        "mean_fit_time",
        "rank_test_precision_macro",
        "rank_test_recall_macro",
        "metric_balanced_accuracy",
        "metric_accuracy",
    ):
        assert isinstance(cv_results[expected_column], np.ndarray), type(
            cv_results[expected_column]
        )

    assert isinstance(cv_results["params"], list), type(cv_results["params"])
    cv_result_items = [
        isinstance(val, npma.MaskedArray)
        for key, val in cv_results.items()
        if key.startswith("param_")
    ]
    assert all(cv_result_items), cv_results.items()

    # Compare the state of the model parameters with the original parameters
    new_params = clone(cls).get_params()
    for param_name, original_value in original_params.items():
        new_value = new_params[param_name]

        # Taken from Sklearn code:
        # We should never change or mutate the internal state of input
        # parameters by default. To check this we use the joblib.hash function
        # that introspects recursively any subobjects to compute a checksum.
        # The only exception to this rule of immutable constructor parameters
        # is possible RandomState instance but in this check we explicitly
        # fixed the random_state params recursively to be integer seeds.
        assert joblib.hash(new_value) == joblib.hash(original_value), (
            "Estimator %s should not change or mutate "
            " the parameter %s from %s to %s during fit."
            % (cls, param_name, original_value, new_value)
        )

    # Comply with https://scikit-learn.org/dev/glossary.html#term-classes
    is_classifier(cls)
    assert hasattr(cls, "classes_")


@pytest.mark.parametrize(
    "estimator_type,dataset_name",
    [(AutoSklearnClassifier, "iris"), (AutoSklearnRegressor, "boston")],
)
def test_leaderboard(
    tmp_dir: str, estimator_type: Type[AutoSklearnEstimator], dataset_name: str
):
    # Comprehensive test tasks a substantial amount of time, manually set if
    # required.
    MAX_COMBO_SIZE_FOR_INCLUDE_PARAM = 3  # [0, len(valid_columns) + 1]
    column_types = AutoSklearnEstimator._leaderboard_columns(num_metrics=1)

    # Create a dict of all possible param values for each param
    # with some invalid one's of the incorrect type
    include_combinations = itertools.chain(
        itertools.combinations(column_types["all"], item_count)
        for item_count in range(1, MAX_COMBO_SIZE_FOR_INCLUDE_PARAM)
    )
    valid_params = {
        "detailed": [True, False],
        "ensemble_only": [True, False],
        "top_k": [-10, 0, 1, 10, "all"],
        "sort_by": [*column_types["all"], "invalid"],
        "sort_order": ["ascending", "descending", "auto", "invalid", None],
        "include": itertools.chain([None, "invalid", "type"], include_combinations),
    }

    # Create a generator of all possible combinations of valid_params
    params_generator = iter(
        dict(zip(valid_params.keys(), param_values))
        for param_values in itertools.product(*valid_params.values())
    )

    X_train, Y_train, _, _ = putil.get_dataset(dataset_name)
    model = estimator_type(
        time_left_for_this_task=30,
        per_run_time_limit=5,
        tmp_folder=os.path.join(tmp_dir, "backend"),
        seed=1,
    )

    model.fit(X_train, Y_train)

    for params in params_generator:
        # Convert from iterator to solid list
        if params["include"] is not None and not isinstance(params["include"], str):
            params["include"] = list(params["include"])

        # Invalid top_k should raise an error, is a positive int or 'all'
        if not (params["top_k"] == "all" or params["top_k"] > 0):
            with pytest.raises(ValueError):
                model.leaderboard(**params)

        # Invalid sort_by column
        elif params["sort_by"] not in column_types["all"]:
            with pytest.raises(ValueError):
                model.leaderboard(**params)

        # Shouldn't accept an invalid sort order
        elif params["sort_order"] not in ["ascending", "descending", "auto"]:
            with pytest.raises(ValueError):
                model.leaderboard(**params)

        # include is single str but not valid
        elif (
            isinstance(params["include"], str)
            and params["include"] not in column_types["all"]
        ):
            with pytest.raises(ValueError):
                model.leaderboard(**params)

        # Crash if include is list but contains invalid column
        elif (
            isinstance(params["include"], list)
            and len(set(params["include"]) - set(column_types["all"])) != 0
        ):
            with pytest.raises(ValueError):
                model.leaderboard(**params)

        # Can't have just model_id, in both single str and list case
        elif params["include"] == "model_id" or params["include"] == ["model_id"]:
            with pytest.raises(ValueError):
                model.leaderboard(**params)

        # Else all valid combinations should be validated
        else:
            leaderboard = model.leaderboard(**params)

            # top_k should never be less than the rows given back
            # It can however be larger
            if isinstance(params["top_k"], int):
                assert params["top_k"] >= len(leaderboard)

            # Check the right columns are present and in the right order
            # The model_id is set as the index, not included in pandas columns
            columns = list(leaderboard.columns)

            def exclude(lst, s):
                return [x for x in lst if x != s]

            if params["include"] is not None:
                # Include with only single str should be the only column
                if isinstance(params["include"], str):
                    assert params["include"] in columns and len(columns) == 1
                # Include as a list should have all the columns without model_id
                else:
                    assert columns == exclude(params["include"], "model_id")
            elif params["detailed"]:
                assert columns == exclude(column_types["detailed"], "model_id")
            else:
                assert columns == exclude(column_types["simple"], "model_id")

            # Ensure that if it's ensemble only
            # Can only check if 'ensemble_weight' is present
            if params["ensemble_only"] and "ensemble_weight" in columns:
                assert all(leaderboard["ensemble_weight"] > 0)


@pytest.mark.parametrize(
    "estimator_type,dataset_name,metrics",
    [
        (
            AutoSklearnClassifier,
            "iris",
            (autosklearn.metrics.accuracy, autosklearn.metrics.balanced_accuracy),
        ),
        (
            AutoSklearnRegressor,
            "boston",
            (autosklearn.metrics.r2, autosklearn.metrics.root_mean_squared_error),
        ),
    ],
)
def test_leaderboard_multi_objective(
    tmp_dir: str,
    estimator_type: Type[AutoSklearnEstimator],
    dataset_name: str,
    metrics: Sequence[autosklearn.metrics.Scorer],
):
    # Comprehensive test tasks a substantial amount of time, manually set if
    # required.
    MAX_COMBO_SIZE_FOR_INCLUDE_PARAM = 3  # [0, len(valid_columns) + 1]
    column_types = AutoSklearnEstimator._leaderboard_columns(num_metrics=2)

    # Create a dict of all possible param values for each param
    # with some invalid one's of the incorrect type
    include_combinations = itertools.chain(
        itertools.combinations(column_types["all"], item_count)
        for item_count in range(1, MAX_COMBO_SIZE_FOR_INCLUDE_PARAM)
    )
    valid_params = {
        "detailed": [True, False],
        "ensemble_only": [True, False],
        "top_k": [-10, 0, 1, 10, "all"],
        "sort_by": [
            "cost",
            "cost_0",
            "cost_1",
            ["cost_1", "cost_0"],
            *column_types["all"],
            "invalid",
        ],
        "sort_order": ["ascending", "descending", "auto", "invalid", None],
        "include": itertools.chain([None, "invalid", "type"], include_combinations),
    }

    # Create a generator of all possible combinations of valid_params
    params_generator = iter(
        dict(zip(valid_params.keys(), param_values))
        for param_values in itertools.product(*valid_params.values())
    )

    X_train, Y_train, _, _ = putil.get_dataset(dataset_name)
    model = estimator_type(
        time_left_for_this_task=30,
        per_run_time_limit=5,
        tmp_folder=os.path.join(tmp_dir, "backend"),
        seed=test.conftest.DEFAULT_SEED,
        metric=metrics,
    )

    model.fit(X_train, Y_train)

    for params in params_generator:
        # Convert from iterator to solid list
        if params["include"] is not None and not isinstance(params["include"], str):
            params["include"] = list(params["include"])

        # Invalid top_k should raise an error, is a positive int or 'all'
        if not (params["top_k"] == "all" or params["top_k"] > 0):
            with pytest.raises(ValueError):
                model.leaderboard(**params)

        # Invalid sort_by column
        elif (
            params["sort_by"] not in column_types["all"]
            and params["sort_by"] != "cost"
            and params["sort_by"] != ["cost_1", "cost_0"]
            and params["sort_by"] not in ["cost_0", "cost_1"]
        ):
            with pytest.raises(ValueError):
                model.leaderboard(**params)

        # Shouldn't accept an invalid sort order
        elif params["sort_order"] not in ["ascending", "descending", "auto"]:
            with pytest.raises(ValueError):
                model.leaderboard(**params)

        # include is single str but not valid
        elif (
            isinstance(params["include"], str)
            and params["include"] not in column_types["all"]
        ):
            with pytest.raises(ValueError):
                model.leaderboard(**params)

        # Crash if include is list but contains invalid column
        elif (
            isinstance(params["include"], list)
            and len(set(params["include"]) - set(column_types["all"])) != 0
        ):
            with pytest.raises(ValueError):
                model.leaderboard(**params)

        # Can't have just model_id, in both single str and list case
        elif params["include"] == "model_id" or params["include"] == ["model_id"]:
            with pytest.raises(ValueError):
                model.leaderboard(**params)

        # Else all valid combinations should be validated
        else:
            leaderboard = model.leaderboard(**params)
            assert "cost" not in leaderboard.columns

            if params["include"] is None:
                assert "cost_0" in leaderboard.columns
                assert "cost_1" in leaderboard.columns
            else:
                for cost_name in ["cost_0", "cost_1"]:
                    if cost_name in params["include"]:
                        assert cost_name in leaderboard.columns

            # top_k should never be less than the rows given back
            # It can however be larger
            if isinstance(params["top_k"], int):
                assert params["top_k"] >= len(leaderboard)

            # Check the right columns are present and in the right order
            # The model_id is set as the index, not included in pandas columns
            columns = list(leaderboard.columns)

            def exclude(lst, s):
                return [x for x in lst if x != s]

            if params["include"] is not None:
                # Include with only single str should be the only column
                if isinstance(params["include"], str):
                    assert params["include"] in columns and len(columns) == 1
                # Include as a list should have all the columns without model_id
                else:
                    assert columns == exclude(params["include"], "model_id")
            elif params["detailed"]:
                assert columns == exclude(column_types["detailed"], "model_id")
            else:
                assert columns == exclude(column_types["simple"], "model_id")

            # Ensure that if it's ensemble only
            # Can only check if 'ensemble_weight' is present
            if params["ensemble_only"] and "ensemble_weight" in columns:
                assert all(leaderboard["ensemble_weight"] > 0)


@pytest.mark.parametrize("estimator", [AutoSklearnRegressor])
@pytest.mark.parametrize("resampling_strategy", ["holdout"])
@pytest.mark.parametrize(
    "X",
    [
        np.asarray(
            [[1.0, 1.0, 1.0]] * 25
            + [[2.0, 2.0, 2.0]] * 25
            + [[3.0, 3.0, 3.0]] * 25
            + [[4.0, 4.0, 4.0]] * 25
        )
    ],
)
@pytest.mark.parametrize(
    "y", [np.asarray([1.0] * 25 + [2.0] * 25 + [3.0] * 25 + [4.0] * 25)]
)
def test_show_models_with_holdout(
    tmp_dir: str,
    dask_client: Client,
    estimator: AutoSklearnEstimator,
    resampling_strategy: str,
    X: np.ndarray,
    y: np.ndarray,
) -> None:
    """
    Parameters
    ----------
    tmp_dir: str
        The temporary directory to use for this test

    dask_client: dask.distributed.Client
         The dask client to use for this test

    estimator: AutoSklearnEstimator
         The estimator to train

    resampling_strategy: str
         The resampling strategy to use

    X: np.ndarray
        The X data to use for this estimator

    y: np.ndarray
         The targets to use for this estimator

    Expects
    -------
    * Expects all the model dictionaries to have ``model_keys``
    * Expects all models to have an auto-sklearn wrapped model ``regressor``
    * Expects all models to have a sklearn wrapped model ``sklearn_regressor``
    * Expects no model to have any ``None`` value
    """

    automl = estimator(
        time_left_for_this_task=60,
        per_run_time_limit=5,
        tmp_folder=os.path.join(tmp_dir, "backend"),
        resampling_strategy=resampling_strategy,
        dask_client=dask_client,
    )
    automl.fit(X, y)

    models = automl.show_models().values()

    model_keys = set(
        [
            "model_id",
            "rank",
            "cost",
            "ensemble_weight",
            "data_preprocessor",
            "feature_preprocessor",
            "regressor",
            "sklearn_regressor",
        ]
    )

    assert all([model_keys == set(model.keys()) for model in models])
    assert all([model["regressor"] for model in models])
    assert all([model["sklearn_regressor"] for model in models])
    assert not any([None in model.values() for model in models])


@pytest.mark.parametrize("estimator", [AutoSklearnClassifier])
@pytest.mark.parametrize("resampling_strategy", ["cv"])
@pytest.mark.parametrize(
    "X", [np.asarray([[1.0, 1.0, 1.0]] * 50 + [[2.0, 2.0, 2.0]] * 50)]
)
@pytest.mark.parametrize("y", [np.asarray([1] * 50 + [2] * 50)])
def test_show_models_with_cv(
    tmp_dir: str,
    dask_client: Client,
    estimator: AutoSklearnEstimator,
    resampling_strategy: str,
    X: np.ndarray,
    y: np.ndarray,
) -> None:
    """
    Parameters
    ----------
    tmp_dir: str
        The temporary directory to use for this test

    dask_client: dask.distributed.Client
         The dask client to use for this test

    estimator: AutoSklearnEstimator
         The estimator to train

    resampling_strategy: str
         The resampling strategy to use

    X: np.ndarray
        The X data to use for this estimator

    y: np.ndarray
         The targets to use for this estimator

    Expects
    -------
    * Expects all the model dictionaries to have `model_keys`
    * Expects no model to have any `None` value
    * Expects all the estimators in a model to have `estimator_keys`
    * Expects all model estimators to have an auto-sklearn wrapped model `classifier`
    * Expects all model estimators to have a sklearn wrapped model `sklearn_classifier`
    * Expects no estimator to have None
    """

    automl = estimator(
        time_left_for_this_task=120,
        per_run_time_limit=5,
        tmp_folder=os.path.join(tmp_dir, "backend"),
        resampling_strategy=resampling_strategy,
        dask_client=dask_client,
    )
    automl.fit(X, y)

    models = automl.show_models().values()

    model_keys = set(
        ["model_id", "rank", "cost", "ensemble_weight", "voting_model", "estimators"]
    )

    estimator_keys = set(
        [
            "data_preprocessor",
            "balancing",
            "feature_preprocessor",
            "classifier",
            "sklearn_classifier",
        ]
    )

    assert all([model_keys == set(model.keys()) for model in models])
    assert not any([None in model.values() for model in models])
    assert all(
        [
            estimator_keys == set(estimator.keys())
            for model in models
            for estimator in model["estimators"]
        ]
    )
    assert all(
        [
            estimator["classifier"]
            for model in models
            for estimator in model["estimators"]
        ]
    )
    assert all(
        [
            estimator["sklearn_classifier"]
            for model in models
            for estimator in model["estimators"]
        ]
    )
    assert not any(
        [
            None in estimator.values()
            for model in models
            for estimator in model["estimators"]
        ]
    )


@unittest.mock.patch("autosklearn.estimators.AutoSklearnEstimator.build_automl")
def test_fit_n_jobs_negative(build_automl_patch):
    n_cores = cpu_count()
    cls = AutoSklearnEstimator(n_jobs=-1)
    cls.fit()
    assert cls._n_jobs == n_cores


def test_get_number_of_available_cores():
    n_cores = cpu_count()
    assert n_cores >= 1, n_cores


@unittest.mock.patch("autosklearn.automl.AutoML.predict")
def test_multiclass_prediction(predict_mock, dask_client):
    predicted_probabilities = [
        [0, 0, 0.99],
        [0, 0.99, 0],
        [0.99, 0, 0],
        [0, 0.99, 0],
        [0, 0, 0.99],
    ]
    predicted_indexes = [2, 1, 0, 1, 2]
    expected_result = ["c", "b", "a", "b", "c"]

    predict_mock.return_value = np.array(predicted_probabilities)

    classifier = AutoMLClassifier(
        time_left_for_this_task=1,
        per_run_time_limit=1,
        dask_client=dask_client,
    )
    classifier.InputValidator = InputValidator(is_classification=True)
    classifier.InputValidator.target_validator.fit(
        pd.DataFrame(expected_result, dtype="category"),
    )
    classifier.InputValidator._is_fitted = True

    actual_result = classifier.predict([None] * len(predicted_indexes))

    np.testing.assert_array_equal(expected_result, actual_result)


@unittest.mock.patch("autosklearn.automl.AutoML.predict")
def test_multilabel_prediction(predict_mock, dask_client):
    predicted_probabilities = [
        [0.99, 0],
        [0.99, 0],
        [0, 0.99],
        [0.99, 0.99],
        [0.99, 0.99],
    ]
    predicted_indexes = np.array([[1, 0], [1, 0], [0, 1], [1, 1], [1, 1]])

    predict_mock.return_value = np.array(predicted_probabilities)

    classifier = AutoMLClassifier(
        time_left_for_this_task=1,
        per_run_time_limit=1,
        dask_client=dask_client,
    )
    classifier.InputValidator = InputValidator(is_classification=True)
    classifier.InputValidator.target_validator.fit(
        pd.DataFrame(predicted_indexes, dtype="int64"),
    )
    classifier.InputValidator._is_fitted = True

    assert (
        classifier.InputValidator.target_validator.type_of_target
        == "multilabel-indicator"
    )

    actual_result = classifier.predict([None] * len(predicted_indexes))

    np.testing.assert_array_equal(predicted_indexes, actual_result)


def test_can_pickle_classifier(tmp_dir, dask_client):
    X_train, Y_train, X_test, Y_test = putil.get_dataset("iris")
    automl = AutoSklearnClassifier(
        time_left_for_this_task=30,
        delete_tmp_folder_after_terminate=False,
        per_run_time_limit=5,
        tmp_folder=os.path.join(tmp_dir, "backend"),
        dask_client=dask_client,
    )

    automl.fit(X_train, Y_train)

    initial_predictions = automl.predict(X_test)
    initial_accuracy = sklearn.metrics.accuracy_score(Y_test, initial_predictions)
    assert initial_accuracy >= 0.75
    assert count_succeses(automl.cv_results_) > 0
    assert includes_train_scores(automl.performance_over_time_.columns) is True
    assert performance_over_time_is_plausible(automl.performance_over_time_) is True

    # Test pickle
    dump_file = os.path.join(tmp_dir, "automl.dump.pkl")

    with open(dump_file, "wb") as f:
        pickle.dump(automl, f)

    with open(dump_file, "rb") as f:
        restored_automl = pickle.load(f)

    restored_predictions = restored_automl.predict(X_test)
    restored_accuracy = sklearn.metrics.accuracy_score(Y_test, restored_predictions)
    assert restored_accuracy >= 0.75
    assert initial_accuracy == restored_accuracy

    # Test joblib
    dump_file = os.path.join(tmp_dir, "automl.dump.joblib")

    joblib.dump(automl, dump_file)

    restored_automl = joblib.load(dump_file)

    restored_predictions = restored_automl.predict(X_test)
    restored_accuracy = sklearn.metrics.accuracy_score(Y_test, restored_predictions)
    assert restored_accuracy >= 0.75
    assert initial_accuracy == restored_accuracy


def test_multilabel(tmp_dir, dask_client):

    X_train, Y_train, X_test, Y_test = putil.get_dataset("iris", make_multilabel=True)
    automl = AutoSklearnClassifier(
        time_left_for_this_task=30,
        per_run_time_limit=5,
        tmp_folder=os.path.join(tmp_dir, "backend"),
        dask_client=dask_client,
    )

    automl.fit(X_train, Y_train)

    predictions = automl.predict(X_test)
    assert predictions.shape == (50, 3), print_debug_information(automl)
    assert count_succeses(automl.cv_results_) > 0, print_debug_information(automl)
    assert includes_train_scores(automl.performance_over_time_.columns) is True
    assert performance_over_time_is_plausible(automl.performance_over_time_) is True

    score = f1_macro(Y_test, predictions)
    assert score >= 0.9, print_debug_information(automl)

    probs = automl.predict_proba(X_train)
    assert np.mean(probs) == pytest.approx(0.33, rel=1e-1)


def test_binary(tmp_dir, dask_client):

    X_train, Y_train, X_test, Y_test = putil.get_dataset("iris", make_binary=True)
    automl = AutoSklearnClassifier(
        time_left_for_this_task=40,
        delete_tmp_folder_after_terminate=False,
        per_run_time_limit=10,
        tmp_folder=os.path.join(tmp_dir, "backend"),
        dask_client=dask_client,
    )

    automl.fit(
        X_train,
        Y_train,
        X_test=X_test,
        y_test=Y_test,
        dataset_name="binary_test_dataset",
    )

    predictions = automl.predict(X_test)
    assert predictions.shape == (50,), print_debug_information(automl)

    score = accuracy(Y_test, predictions)
    assert score > 0.9, print_debug_information(automl)
    assert count_succeses(automl.cv_results_) > 0, print_debug_information(automl)
    assert includes_all_scores(automl.performance_over_time_.columns) is True
    assert performance_over_time_is_plausible(automl.performance_over_time_) is True


def test_classification_pandas_support(tmp_dir, dask_client):

    X, y = sklearn.datasets.fetch_openml(
        data_id=2,  # cat/num dataset
        return_X_y=True,
        as_frame=True,
    )

    # Drop NAN!!
    X = X.dropna(axis="columns")

    # This test only make sense if input is dataframe
    assert isinstance(X, pd.DataFrame)
    assert isinstance(y, pd.Series)
    automl = AutoSklearnClassifier(
        time_left_for_this_task=30,
        per_run_time_limit=5,
        exclude={"classifier": ["libsvm_svc"]},
        dask_client=dask_client,
        seed=5,
        tmp_folder=os.path.join(tmp_dir, "backend"),
    )

    automl.fit(X, y)

    # Make sure that at least better than random.
    # We use same X_train==X_test to test code quality
    assert automl.score(X, y) > 0.555, print_debug_information(automl)

    automl.refit(X, y)

    # Make sure that at least better than random.
    # accuracy in sklearn needs valid data
    # It should be 0.555 as the dataset is unbalanced.
    prediction = automl.predict(X)
    assert accuracy(y, prediction) > 0.555
    assert count_succeses(automl.cv_results_) > 0
    assert includes_train_scores(automl.performance_over_time_.columns) is True
    assert performance_over_time_is_plausible(automl.performance_over_time_) is True


def test_regression(tmp_dir, dask_client):

    X_train, Y_train, X_test, Y_test = putil.get_dataset("boston")
    automl = AutoSklearnRegressor(
        time_left_for_this_task=30,
        per_run_time_limit=5,
        tmp_folder=os.path.join(tmp_dir, "backend"),
        dask_client=dask_client,
    )

    automl.fit(X_train, Y_train)

    predictions = automl.predict(X_test)
    assert predictions.shape == (356,)
    score = mean_squared_error(Y_test, predictions)

    # On average np.sqrt(30) away from the target -> ~5.5 on average
    # Results with select rates drops avg score to a range of -32.40 to -37,
    # on 30 seconds constraint.
    # With more time_left_for_this_task this is no longer an issue
    assert score >= -37, print_debug_information(automl)
    assert count_succeses(automl.cv_results_) > 0
    assert includes_train_scores(automl.performance_over_time_.columns) is True
    assert performance_over_time_is_plausible(automl.performance_over_time_) is True


def test_cv_regression(tmp_dir, dask_client):
    """
    Makes sure that when using a cv strategy, we are able to fit
    a regressor
    """

    X_train, Y_train, X_test, Y_test = putil.get_dataset(
        "boston", train_size_maximum=300
    )
    automl = AutoSklearnRegressor(
        time_left_for_this_task=60,
        per_run_time_limit=10,
        resampling_strategy="cv",
        tmp_folder=os.path.join(tmp_dir, "backend"),
        dask_client=dask_client,
    )

    automl.fit(X_train, Y_train)

    predictions = automl.predict(X_test)
    assert predictions.shape == (206,)
    score = r2(Y_test, predictions)
    assert score >= 0.1, print_debug_information(automl)
    assert count_succeses(automl.cv_results_) > 0, print_debug_information(automl)
    assert includes_train_scores(automl.performance_over_time_.columns) is True
    assert performance_over_time_is_plausible(automl.performance_over_time_) is True


def test_regression_pandas_support(tmp_dir, dask_client):

    X, y = sklearn.datasets.fetch_openml(
        data_id=41514,  # diabetes
        return_X_y=True,
        as_frame=True,
    )
    # This test only make sense if input is dataframe
    assert isinstance(X, pd.DataFrame)
    assert isinstance(y, pd.Series)
    automl = AutoSklearnRegressor(
        time_left_for_this_task=40,
        per_run_time_limit=5,
        dask_client=dask_client,
        tmp_folder=os.path.join(tmp_dir, "backend"),
    )

    # Make sure we error out because y is not encoded
    automl.fit(X, y)

    # Make sure that at least better than random.
    # We use same X_train==X_test to test code quality
    assert automl.score(X, y) >= 0.5, print_debug_information(automl)

    automl.refit(X, y)

    # Make sure that at least better than random.
    assert r2(y, automl.predict(X)) > 0.5, print_debug_information(automl)
    assert count_succeses(automl.cv_results_) > 0, print_debug_information(automl)
    assert includes_train_scores(automl.performance_over_time_.columns) is True
    assert performance_over_time_is_plausible(automl.performance_over_time_) is True


def test_autosklearn_classification_methods_returns_self(dask_client):
    """Currently this method only tests that the methods of AutoSklearnClassifier
    is able to fit using fit(), fit_ensemble() and refit()
    """
    X_train, y_train, X_test, y_test = putil.get_dataset("iris")
    automl = AutoSklearnClassifier(
        time_left_for_this_task=60,
        delete_tmp_folder_after_terminate=False,
        per_run_time_limit=10,
        dask_client=dask_client,
        exclude={"feature_preprocessor": ["fast_ica"]},
    )

    automl_fitted = automl.fit(X_train, y_train)

    assert automl is automl_fitted

    automl_ensemble_fitted = automl.fit_ensemble(
        y_train, ensemble_kwargs={"ensemble_size": 5}
    )
    assert automl is automl_ensemble_fitted

    automl_refitted = automl.refit(X_train.copy(), y_train.copy())

    assert automl is automl_refitted


# Currently this class only tests that the methods of AutoSklearnRegressor
# that should return self actually return self.
def test_autosklearn_regression_methods_returns_self(dask_client):
    X_train, y_train, X_test, y_test = putil.get_dataset("boston")
    automl = AutoSklearnRegressor(
        time_left_for_this_task=30,
        delete_tmp_folder_after_terminate=False,
        per_run_time_limit=5,
        dask_client=dask_client,
    )

    automl_fitted = automl.fit(X_train, y_train)
    assert automl is automl_fitted

    automl_ensemble_fitted = automl.fit_ensemble(
        y_train, ensemble_kwargs={"ensemble_size": 5}
    )
    assert automl is automl_ensemble_fitted

    automl_refitted = automl.refit(X_train.copy(), y_train.copy())
    assert automl is automl_refitted


def test_autosklearn2_classification_methods_returns_self(dask_client):
    X_train, y_train, X_test, y_test = putil.get_dataset("iris")
    automl = AutoSklearn2Classifier(
        time_left_for_this_task=60,
        delete_tmp_folder_after_terminate=False,
        dask_client=dask_client,
    )

    automl_fitted = automl.fit(X_train, y_train)

    assert automl is automl_fitted

    automl_ensemble_fitted = automl.fit_ensemble(
        y_train, ensemble_kwargs={"ensemble_size": 5}
    )
    assert automl is automl_ensemble_fitted

    automl_refitted = automl.refit(X_train.copy(), y_train.copy())

    assert automl is automl_refitted

    predictions = automl_fitted.predict(X_test)
    assert (
        sklearn.metrics.accuracy_score(y_test, predictions) >= 4 / 5
    ), print_debug_information(automl)

    pickle.dumps(automl_fitted)


def test_autosklearn2_classification_methods_returns_self_sparse(dask_client):
    X_train, y_train, X_test, y_test = putil.get_dataset(
        "breast_cancer", make_sparse=True
    )
    automl = AutoSklearn2Classifier(
        time_left_for_this_task=60,
        delete_tmp_folder_after_terminate=False,
        dask_client=dask_client,
    )

    automl_fitted = automl.fit(X_train, y_train)

    assert automl is automl_fitted

    automl_ensemble_fitted = automl.fit_ensemble(
        y_train, ensemble_kwargs={"ensemble_size": 5}
    )
    assert automl is automl_ensemble_fitted

    automl_refitted = automl.refit(X_train.copy(), y_train.copy())

    assert automl is automl_refitted

    predictions = automl_fitted.predict(X_test)
    assert (
        sklearn.metrics.accuracy_score(y_test, predictions) >= 4 / 5
    ), print_debug_information(automl)

    assert "boosting" not in str(automl.get_configuration_space(X=X_train, y=y_train))

    pickle.dumps(automl_fitted)


@pytest.mark.parametrize(
    "class_", [AutoSklearnClassifier, AutoSklearnRegressor, AutoSklearn2Classifier]
)
def test_check_estimator_signature(class_):
    # Make sure signature is store in self
    expected_subclass = (
        ClassifierMixin if "Classifier" in str(class_) else RegressorMixin
    )
    assert issubclass(class_, expected_subclass)
    estimator = class_()
    for expected in list(inspect.signature(class_).parameters):
        assert hasattr(estimator, expected)


@pytest.mark.parametrize(
    "selector_path",
    [
        None,  # No XDG_CACHE_HOME provided
        "/",  # XDG_CACHE_HOME has no permission
        tempfile.gettempdir(),  # in the user cache
    ],
)
def test_selector_file_askl2_can_be_created(selector_path):

    with unittest.mock.patch("os.environ.get") as mock_foo:
        mock_foo.return_value = selector_path
        if selector_path is not None and not os.access(selector_path, os.W_OK):
            with pytest.raises(PermissionError):
                importlib.reload(autosklearn.experimental.askl2)
                automl = AutoSklearn2Classifier(
                    time_left_for_this_task=60, delete_tmp_folder_after_terminate=False
                )

        else:
            importlib.reload(autosklearn.experimental.askl2)
            automl = AutoSklearn2Classifier(
                time_left_for_this_task=60, delete_tmp_folder_after_terminate=False
            )
            for metric in automl.selector_metrics:
                assert os.path.exists(automl.selector_files[metric.name])
                if selector_path is None or not os.access(selector_path, os.W_OK):
                    # We default to home in worst case
                    assert os.path.expanduser("~") in str(
                        automl.selector_files[metric.name]
                    )
                else:
                    # a dir provided via XDG_CACHE_HOME
                    assert selector_path in str(automl.selector_files[metric.name])
    # Re import it at the end so we do not affect other test
    importlib.reload(autosklearn.experimental.askl2)


@pytest.mark.parametrize(
    "metric",
    [metric for metric in autosklearn.experimental.askl2.selector_metrics],
)
def test_askl2_fits_selector_for_given_metrics_at_init(tmp_path, metric):

    assert tmp_path.is_dir()
    assert len(list(tmp_path.iterdir())) == 0
    temp_dir = str(tmp_path)

    with unittest.mock.patch("os.environ.get") as mock_foo:
        mock_foo.return_value = temp_dir
        automl = AutoSklearn2Classifier(
            time_left_for_this_task=60,
            delete_tmp_folder_after_terminate=False,
            metric=metric,
        )
        assert (
            len(automl.selector_files) == 1
        )  # only one selector file should have been created
        assert os.path.exists(
            str(automl.selector_files[metric.name])
        )  # check if the path exists

        # check if selector is retrained when
        # another object with the same metric is created
        automl_1 = AutoSklearn2Classifier(
            time_left_for_this_task=60,
            delete_tmp_folder_after_terminate=False,
            metric=metric,
        )
        assert (
            len(automl_1.selector_files) == 1
        )  # only one selector file should have been created
        assert os.path.exists(
            str(automl_1.selector_files[metric.name])
        )  # check if the path exists
        assert not automl_1.required_training


def test_askl2_fit_when_no_metric_specified(tmp_path):

    assert tmp_path.is_dir()
    assert len(list(tmp_path.iterdir())) == 0
    temp_dir = str(tmp_path)

    with unittest.mock.patch("os.environ.get") as mock_foo:
        mock_foo.return_value = temp_dir
        automl = AutoSklearn2Classifier(
            time_left_for_this_task=60, delete_tmp_folder_after_terminate=False
        )
        assert len(automl.selector_files) == 3
        for metric in automl.selector_metrics:
            assert os.path.exists(str(automl.selector_files[metric.name]))


def test_check_askl2_same_arguments_as_askl() -> None:
    """Check the asklearn2 has the same args as asklearn1

    Expects
    -------
    * The set of arguments for AutoSklearnClassifier is the same as
        AutoSklearn2Classifier except for a few expected arugments.
    """
    autosklearn1_classifier_args = set(
        inspect.getfullargspec(AutoSklearnEstimator.__init__).args
    )
    autosklearn2_classifier_args = set(
        inspect.getfullargspec(AutoSklearn2Classifier.__init__).args
    )
    extra_arguments = autosklearn1_classifier_args - autosklearn2_classifier_args

    expected_extra_args = set(
        [
            "exclude",
            "include",
            "resampling_strategy_arguments",
            "get_smac_object_callback",
            "initial_configurations_via_metalearning",
            "resampling_strategy",
            "metadata_directory",
            "get_trials_callback",
        ]
    )
    unexpected_args = extra_arguments - expected_extra_args

    assert len(unexpected_args) == 0, unexpected_args


@pytest.mark.parametrize("task_type", ["classification", "regression"])
@pytest.mark.parametrize("resampling_strategy", ["test", "cv", "holdout"])
@pytest.mark.parametrize("disable_file_output", [True, False])
def test_fit_pipeline(dask_client, task_type, resampling_strategy, disable_file_output):
    """
    Tests that we can query the configuration space, and from the default configuration
    space, fit a classification pipeline with an acceptable score
    """
    X_train, y_train, X_test, y_test = putil.get_dataset(
        "iris" if task_type == "classification" else "boston"
    )
    estimator = (
        AutoSklearnClassifier if task_type == "classification" else AutoSklearnRegressor
    )
    seed = 3
    if task_type == "classification":
        include = {"classifier": ["random_forest"]}
    else:
        include = {"regressor": ["random_forest"]}
    automl = estimator(
        delete_tmp_folder_after_terminate=False,
        time_left_for_this_task=120,
        # Time left for task plays no role
        # only per run time limit
        per_run_time_limit=30,
        dask_client=dask_client,
        include=include,
        seed=seed,
        # We cannot get the configuration space with 'test' not fit with it
        resampling_strategy=resampling_strategy
        if resampling_strategy != "test"
        else "holdout",
    )
    config = automl.get_configuration_space(
        X_train,
        y_train,
        X_test=X_test,
        y_test=y_test,
    ).get_default_configuration()

    pipeline, run_info, run_value = automl.fit_pipeline(
        X=X_train,
        y=y_train,
        config=config,
        X_test=X_test,
        y_test=y_test,
        disable_file_output=disable_file_output,
        resampling_strategy=resampling_strategy,
    )

    assert isinstance(run_info.config, Configuration)
    assert run_info.cutoff == 30
    assert run_value.status == StatusType.SUCCESS, f"{run_info}->{run_value}"
    # We should produce a decent result
    assert run_value.cost < 0.2

    # Make sure that the pipeline can be pickled
    dump_file = os.path.join(tempfile.gettempdir(), "automl.dump.pkl")
    with open(dump_file, "wb") as f:
        pickle.dump(pipeline, f)

    if resampling_strategy == "test" or disable_file_output:
        # We do not produce a pipeline in 'test'
        assert pipeline is None
    elif resampling_strategy == "cv":
        # We should have fitted a Voting estimator
        assert hasattr(pipeline, "estimators_")
    else:
        # We should have fitted a pipeline with named_steps
        assert hasattr(pipeline, "named_steps")
        assert "RandomForest" in pipeline.steps[-1][-1].choice.__class__.__name__

    # Num run should be 2, as 1 is for dummy classifier and we have not launch
    # another pipeline
    num_run = 2

    # Check the re-sampling strategy
    num_run_dir = automl.automl_._backend.get_numrun_directory(
        seed, num_run, budget=0.0
    )
    cv_model_path = os.path.join(
        num_run_dir,
        automl.automl_._backend.get_cv_model_filename(seed, num_run, budget=0.0),
    )
    model_path = os.path.join(
        num_run_dir,
        automl.automl_._backend.get_model_filename(seed, num_run, budget=0.0),
    )
    if resampling_strategy == "test" or disable_file_output:
        # No file output is expected
        assert not os.path.exists(num_run_dir)
    else:
        # We expect the model path always
        # And the cv model only on 'cv'
        assert os.path.exists(model_path)
        if resampling_strategy == "cv":
            assert os.path.exists(cv_model_path)
        elif resampling_strategy == "holdout":
            assert not os.path.exists(cv_model_path)


@pytest.mark.parametrize("data_type", ["pandas", "numpy"])
@pytest.mark.parametrize("include_categorical", [True, False])
def test_pass_categorical_and_numeric_columns_to_pipeline(
    dask_client, data_type, include_categorical
):
    # Prepare the training data
    X, y = sklearn.datasets.make_classification(random_state=0)
    X = cast(np.ndarray, X)

    n_features = X.shape[1]

    # If categorical, insert a row of 'categorical' '0's at last col
    if include_categorical:
        X = np.insert(X, n_features, values=0, axis=1)

    if data_type == "pandas":
        X = pd.DataFrame(X)
        y = pd.DataFrame(y, dtype="category")

        # Set the last column to categorical
        if include_categorical:
            X.loc[:, n_features] = X.loc[:, n_features].astype("category")

    # Specify the feature_types
    if data_type == "numpy" and include_categorical:
        feat_type = ["numerical"] * n_features + ["categorical"]
    else:
        feat_type = None

    # Create the splits
    X_train, X_test, y_train, y_test = model_selection.train_test_split(
        X, y, test_size=0.5, random_state=3
    )

    # Create Estimator
    # Time left for task plays no role for fit_pipeline
    automl = AutoSklearnClassifier(
        delete_tmp_folder_after_terminate=False,
        time_left_for_this_task=120,
        per_run_time_limit=30,
        seed=0,
        dask_client=dask_client,
        include={"classifier": ["random_forest"]},
    )

    config_space = automl.get_configuration_space(
        X_train,
        y_train,
        X_test=X_test,
        y_test=y_test,
        feat_type=feat_type,
    )
    config = config_space.get_default_configuration()

    pipeline, _, run_value = automl.fit_pipeline(
        X=X_train,
        y=y_train,
        X_test=X_test,
        y_test=y_test,
        config=config,
        feat_type=feat_type,
    )

    assert pipeline is not None, "Expected a pipeline from automl.fit_pipeline"

    feature_validator = automl.automl_.InputValidator.feature_validator  # type: ignore
    transformed_X_test = feature_validator.transform(X_test)
    predictions = pipeline.predict(transformed_X_test)

    # We should produce a half decent result
    assert run_value.cost < 0.40, f"Run value:\n {run_value}"

    # Outputs should be the correct length
    assert np.shape(predictions)[0] == np.shape(y_test)[0]

    n_columns = np.shape(X)[1]

    if include_categorical:
        expected_feat_types = {
            i: feature_type
            for i, feature_type in enumerate(
                ["numerical"] * (n_columns - 1) + ["categorical"]
            )
        }

    else:
        expected_feat_types = {
            i: feature_type for i, feature_type in enumerate(["numerical"] * n_columns)
        }

    pipeline_feat_types = pipeline.named_steps["data_preprocessor"].choice.feat_type
    assert expected_feat_types == pipeline_feat_types


@pytest.mark.parametrize("as_frame", [True, False])
def test_autosklearn_anneal(as_frame):
    """
    This test makes sure that anneal dataset can be fitted and scored.
    This dataset is quite complex, with NaN, categorical and numerical columns
    so is a good testcase for unit-testing
    """
    X, y = sklearn.datasets.fetch_openml(data_id=2, return_X_y=True, as_frame=as_frame)
    automl = AutoSklearnClassifier(
        time_left_for_this_task=60,
        delete_tmp_folder_after_terminate=False,
        initial_configurations_via_metalearning=0,
        smac_scenario_args={"runcount_limit": 6},
        resampling_strategy="holdout-iterative-fit",
    )

    if as_frame:
        # Let autosklearn calculate the feat types
        automl_fitted = automl.fit(X, y)

    else:
        X_, y_ = sklearn.datasets.fetch_openml(
            data_id=2, return_X_y=True, as_frame=True
        )
        feat_type = [
            "categorical" if X_[col].dtype.name == "category" else "numerical"
            for col in X_.columns
        ]

        automl_fitted = automl.fit(X, y, feat_type=feat_type)

    assert automl is automl_fitted

    automl_ensemble_fitted = automl.fit_ensemble(
        y, ensemble_kwargs={"ensemble_size": 5}
    )
    assert automl is automl_ensemble_fitted

    # We want to make sure we can learn from this data.
    # This is a test to make sure the data format (numpy/pandas)
    # can be used in a meaningful way -- not meant for generalization,
    # hence we use the train dataset
    assert automl_fitted.score(X, y) > 0.75


@pytest.mark.parametrize(
    "dataset_compression", [False, True, {"memory_allocation": 0.2}]
)
def test_param_dataset_compression(dataset_compression: Union[bool, Dict[str, Any]]):
    """We expect this does not get parsed and modified until it gets to the AutoML
    class, In the meantime, it's value remains whatever was passed in.

    Expects
    -------
    * The private attribute should be set to what was passed
    """
    model = AutoSklearnClassifier(dataset_compression=dataset_compression)

    assert model.dataset_compression == dataset_compression


def test_ensemble_default_resolves():
    model = AutoSklearnClassifier(
        ensemble_class="default",
        metric=accuracy,
    )
    assert model.ensemble_class == EnsembleSelection
    model = AutoSklearnClassifier(
        ensemble_class="default",
        metric=[accuracy, f1_macro],
    )
    assert model.ensemble_class == MultiObjectiveDummyEnsemble



from __future__ import annotations

from typing import Any, Callable

import copy
import math
import pickle
import sys
from pathlib import Path

import numpy as np

from autosklearn.automl import AutoML
from autosklearn.automl_common.common.utils.backend import Backend
from autosklearn.constants import BINARY_CLASSIFICATION
from autosklearn.data.xy_data_manager import XYDataManager
from autosklearn.ensemble_building import EnsembleBuilder, EnsembleBuilderManager, Run
from autosklearn.metrics import Scorer, accuracy

from pytest_cases import fixture

from test.conftest import DEFAULT_SEED


@fixture
def make_run(tmp_path: Path) -> Callable[..., Run]:
    def _make(
        id: int | None = None,
        dummy: bool = False,
        backend: Backend | None = None,
        seed: int = DEFAULT_SEED,
        modified: bool = False,
        budget: float = 0.0,
        loss: float | None = None,
        losses: dict[str, float] | list[float] | None = None,
        model_size: int | None = None,
        mem_usage: float | None = None,
        predictions: str | list[str] | dict[str, np.ndarray] | None = "ensemble",
    ) -> Run:
        if loss is not None and losses is not None:
            raise ValueError("Can only specify either `loss` or `losses`")

        if isinstance(loss, dict):
            raise ValueError("Please use `losses` for dict of losses")

        if dummy:
            assert id is None
            id = 1
            if loss is None and losses is None:
                losses = {"metric_0": 50_000}

        if id is None:
            id = np.random.randint(sys.maxsize)

        model_id = f"{seed}_{id}_{budget}"

        # Use this backend to set things up
        if backend is not None:
            runsdir = Path(backend.get_runs_directory())
        else:
            runsdir = tmp_path

        dir = runsdir / model_id

        if not dir.exists():
            dir.mkdir()

        # Populate if None
        if isinstance(predictions, str):
            predictions = [predictions]

        # Convert to dict
        if isinstance(predictions, list):
            preds = np.asarray([[0]])
            predictions = {kind: preds for kind in predictions}

        # Write them
        if isinstance(predictions, dict):
            for kind, val in predictions.items():
                fname = f"predictions_{kind}_{seed}_{id}_{budget}.npy"
                with (dir / fname).open("wb") as f:
                    np.save(f, val)

        run = Run(dir)

        if modified:
            assert predictions is not None, "Can only modify if predictions"
            for k, v in run.recorded_mtimes.items():
                run.recorded_mtimes[k] = v + 1e-4

        if loss is not None:
            losses = [loss]

        if isinstance(losses, list):
            losses = {f"metric_{i}": loss for i, loss in enumerate(losses)}

        if isinstance(losses, dict):
            run.losses = losses

        if mem_usage is not None:
            run._mem_usage = mem_usage

        # MB
        if model_size is not None:
            n_bytes = int(model_size * math.pow(1024, 2))
            model_path = dir / f"{seed}.{id}.{budget}.model"
            with model_path.open("wb") as f:
                f.write(bytearray(n_bytes))

        return run

    return _make


@fixture
def make_ensemble_builder(
    make_backend: Callable[..., Backend],
    make_sklearn_dataset: Callable[..., XYDataManager],
) -> Callable[..., EnsembleBuilder]:
    def _make(
        *,
        automl: AutoML | None = None,
        previous_candidates: list[Run] | None = None,
        backend: Backend | None = None,
        dataset_name: str = "TEST",
        task_type: int = BINARY_CLASSIFICATION,
        metrics: Scorer = copy.deepcopy([accuracy]),
        **kwargs: Any,
    ) -> EnsembleBuilder:

        if automl:
            backend = automl._backend
            dataset_name = automl._dataset_name
            task_type = automl._task
            metrics = automl._metrics
            kwargs = {
                "ensemble_class": automl._ensemble_class,
                "ensemble_kwargs": automl._ensemble_kwargs,
                "ensemble_nbest": automl._ensemble_nbest,
                "max_models_on_disc": automl._max_models_on_disc,
                "precision": automl.precision,
                "read_at_most": automl._read_at_most,
                "memory_limit": automl._memory_limit,
                "logger_port": automl._logger_port,
            }

        if backend is None:
            backend = make_backend()

        # If there's no datamanager, just try populate it with some generic one,
        if not Path(backend._get_datamanager_pickle_filename()).exists():
            datamanager = make_sklearn_dataset(
                name="breast_cancer",
                task=BINARY_CLASSIFICATION,
                feat_type="numerical",  # They're all numerical
                as_datamanager=True,
            )
            backend.save_datamanager(datamanager)

            # Annoyingly, some places use datamanger, some places use the file
            # Hence, we take the y_train of the datamanager and use that as the
            # the targets
            if "Y_train" in datamanager.data:
                backend.save_additional_data(
                    datamanager.data["Y_train"], what="targets_ensemble"
                )
            if "X_train" in datamanager.data:
                backend.save_additional_data(
                    datamanager.data["X_train"], what="input_ensemble"
                )

        builder = EnsembleBuilder(
            backend=backend,
            dataset_name=dataset_name,
            task_type=task_type,
            metrics=metrics,
            **kwargs,
        )

        if previous_candidates is not None:
            with builder.previous_candidates_path.open("wb") as f:
                pickle.dump({run.id: run for run in previous_candidates}, f)

        return builder

    return _make


@fixture
def make_ensemble_builder_manager(
    make_backend: Callable[..., Backend],
    make_sklearn_dataset: Callable[..., XYDataManager],
) -> Callable[..., EnsembleBuilderManager]:
    """Use `make_run` to create runs for this manager

    .. code:: python

        def test_x(make_run, make_ensemble_builder_manager):
            manager = make_ensemble_builder(...)

            # Will use the backend to place runs correctly
            runs = make_run(predictions={"ensemble": ...}, backend=manager.backend)

            # ... test stuff


    """

    def _make(
        *,
        backend: Backend | None = None,
        dataset_name: str = "TEST",
        task: int = BINARY_CLASSIFICATION,
        metric: Scorer = accuracy,
        random_state: int | np.random.RandomState | None = DEFAULT_SEED,
        **kwargs: Any,
    ) -> EnsembleBuilderManager:
        if backend is None:
            backend = make_backend()

        if not Path(backend._get_datamanager_pickle_filename()).exists():
            datamanager = make_sklearn_dataset(
                name="breast_cancer",
                task=BINARY_CLASSIFICATION,
                feat_type="numerical",  # They're all numerical
                as_datamanager=True,
            )
            backend.save_datamanager(datamanager)

            # Annoyingly, some places use datamanger, some places use the file
            # Hence, we take the y_train of the datamanager and use that as the
            # the targets
            if "Y_train" in datamanager.data:
                backend.save_additional_data(
                    datamanager.data["Y_train"],
                    what="targets_ensemble",
                )

        return EnsembleBuilderManager(
            backend=backend,
            dataset_name=dataset_name,
            task=task,
            metric=metric,
            random_state=random_state,
            **kwargs,
        )

    return _make



from __future__ import annotations

from typing import Callable, Collection, Optional, Union

import numpy as np
from sklearn.ensemble import VotingClassifier, VotingRegressor

from autosklearn.data.validation import (
    SUPPORTED_FEAT_TYPES,
    SUPPORTED_TARGET_TYPES,
    InputValidator,
)
from autosklearn.evaluation.abstract_evaluator import (
    MyDummyClassifier,
    MyDummyRegressor,
)
from autosklearn.pipeline.components.base import (
    AutoSklearnClassificationAlgorithm,
    AutoSklearnRegressionAlgorithm,
)

from pytest_cases import fixture

from test.conftest import DEFAULT_SEED


@fixture
def make_voting_classifier() -> Callable[..., VotingClassifier]:
    """
    Parameters
    ----------
    X: Optional[SUPPORTED_FEAT_TYPES] = None
        The X data to fit models on, if None, no fitting occurs

    y: Optional[SUPPORTED_FEAT_TYPES] = None
        The y data to fit models on, if None, no fitting occurs

    models: Optional[Collection[AutoSklearnClassificationAlgorithm]] = None
        Any collection of algorithms to use, if None, DummyClassifiers are used
    """

    def _make(
        X: Optional[SUPPORTED_FEAT_TYPES] = None,
        y: Optional[SUPPORTED_TARGET_TYPES] = None,
        models: Optional[Collection[AutoSklearnClassificationAlgorithm]] = None,
        seed: Union[int, None, np.random.RandomState] = DEFAULT_SEED,
    ) -> VotingClassifier:
        assert not (X is None) ^ (y is None)
        if not models:
            validator = InputValidator(is_classification=True).fit(X, y)
            models = [
                MyDummyClassifier(
                    feat_type=validator.feature_validator.feat_type,
                    config=1,
                    random_state=seed,
                )
                for _ in range(5)
            ]

        if X is not None:
            for model in models:
                model.fit(X, y)

        voter = VotingClassifier(estimators=None, voting="soft")
        voter.estimators_ = models
        return voter

    return _make


@fixture
def make_voting_regressor() -> Callable[..., VotingRegressor]:
    """
    Parameters
    ----------
    X: Optional[SUPPORTED_FEAT_TYPES] = None
        The X data to fit models on, if None, no fitting occurs

    y: Optional[SUPPORTED_FEAT_TYPES] = None
        The y data to fit models on, if None, no fitting occurs

    models: Optional[Collection[AutoSklearnRegressionAlgorithm]] = None
        Any collection of algorithms to use, if None, DummyRegressors are used
    """

    def _make(
        X: Optional[SUPPORTED_FEAT_TYPES] = None,
        y: Optional[SUPPORTED_TARGET_TYPES] = None,
        models: Optional[Collection[AutoSklearnRegressionAlgorithm]] = None,
        seed: Union[int, None, np.random.RandomState] = DEFAULT_SEED,
    ) -> VotingRegressor:
        assert not (X is None) ^ (y is None)

        if not models:
            validator = InputValidator(is_classification=False).fit(X, y)
            models = [
                MyDummyRegressor(
                    feat_type=validator.feature_validator.feat_type,
                    config=1,
                    random_state=seed,
                )
                for _ in range(5)
            ]

        if X is not None:
            for model in models:
                model.fit(X, y)

        voter = VotingRegressor(estimators=None)
        voter.estimators_ = models
        return voter

    return _make



import logging
import logging.config
import os
import tempfile

import yaml

from autosklearn.util import logging_

import unittest


class LoggingTest(unittest.TestCase):
    def test_setup_logger(self):
        # Test that setup_logger function correctly configures the logger
        # according to the given dictionary, and uses the default
        # logging.yaml file if logging_config is not specified.

        with open(
            os.path.join(os.path.dirname(__file__), "example_config.yaml"), "r"
        ) as fh:
            example_config = yaml.safe_load(fh)

        # Configure logger with example_config.yaml.
        logging_.setup_logger(
            logging_config=example_config, output_dir=tempfile.gettempdir()
        )

        # example_config sets the root logger's level to CRITICAL,
        # which corresponds to 50.
        self.assertEqual(logging.getLogger().getEffectiveLevel(), 50)

        # This time use the default configuration.
        logging_.setup_logger(logging_config=None, output_dir=tempfile.gettempdir())

        # default config sets the root logger's level to DEBUG,
        # which corresponds to 10.
        self.assertEqual(logging.getLogger().getEffectiveLevel(), 10)

        # Make sure we log to the desired directory
        logging_.setup_logger(output_dir=os.path.dirname(__file__), filename="test.log")
        logger = logging.getLogger()
        logger.info("test_setup_logger")

        with open(os.path.join(os.path.dirname(__file__), "test.log")) as fh:
            self.assertIn("test_setup_logger", "".join(fh.readlines()))
        os.remove(os.path.join(os.path.dirname(__file__), "test.log"))



from typing import Dict, Optional, Tuple, Union

import numpy as np
import scipy.sparse
from ConfigSpace.configuration_space import ConfigurationSpace
from sklearn.preprocessing import OrdinalEncoder

from autosklearn.askl_typing import FEAT_TYPE_TYPE
from autosklearn.pipeline.base import DATASET_PROPERTIES_TYPE, PIPELINE_DATA_DTYPE
from autosklearn.pipeline.components.base import AutoSklearnPreprocessingAlgorithm
from autosklearn.pipeline.constants import DENSE, INPUT, SPARSE, UNSIGNED_DATA


class OrdinalEncoding(AutoSklearnPreprocessingAlgorithm):
    def __init__(
        self, random_state: Optional[Union[int, np.random.RandomState]] = None
    ) -> None:
        self.random_state = random_state

    def fit(
        self, X: PIPELINE_DATA_DTYPE, y: Optional[PIPELINE_DATA_DTYPE] = None
    ) -> "OrdinalEncoding":
        if not scipy.sparse.issparse(X):
            self.preprocessor = OrdinalEncoder(
                categories="auto",
                handle_unknown="use_encoded_value",
                unknown_value=-1,
            )
            self.preprocessor.fit(X, y)
            return self
        else:
            # TODO sparse_encoding of negative labels
            #
            #   The next step in the pipeline relies on positive labels
            #   Given a categorical column [[0], [-1]], the next step will fail
            #   unless we can fix this encoding
            return self

    def transform(self, X: PIPELINE_DATA_DTYPE) -> PIPELINE_DATA_DTYPE:
        if scipy.sparse.issparse(X):
            # Sparse data should be float dtype, which means we do not need
            # to further encode it.
            return X
        if self.preprocessor is None:
            raise NotImplementedError()
        # Notice we are shifting the unseen categories during fit to 1
        # from -1, 0, ... to 0,..., cat + 1
        # This is done because Category shift requires non negative integers
        # Consider removing this if that step is removed
        return self.preprocessor.transform(X) + 1

    @staticmethod
    def get_properties(
        dataset_properties: Optional[DATASET_PROPERTIES_TYPE] = None,
    ) -> Dict[str, Optional[Union[str, int, bool, Tuple]]]:
        return {
            "shortname": "OrdinalEncoder",
            "name": "Ordinal Encoder",
            "handles_regression": True,
            "handles_classification": True,
            "handles_multiclass": True,
            "handles_multilabel": True,
            "handles_multioutput": True,
            # TODO find out of this is right!
            "handles_sparse": True,
            "handles_dense": True,
            "input": (DENSE, SPARSE, UNSIGNED_DATA),
            "output": (INPUT,),
        }

    @staticmethod
    def get_hyperparameter_search_space(
        feat_type: Optional[FEAT_TYPE_TYPE] = None,
        dataset_properties: Optional[DATASET_PROPERTIES_TYPE] = None,
    ) -> ConfigurationSpace:
        return ConfigurationSpace()



import sklearn.naive_bayes

from autosklearn.pipeline.components.classification.bernoulli_nb import BernoulliNB

from .test_base import BaseClassificationComponentTest


class BernoulliNBComponentTest(BaseClassificationComponentTest):

    __test__ = True

    res = dict()
    res["default_iris"] = 0.26
    res["iris_n_calls"] = None
    res["default_iris_iterative"] = 0.26
    res["default_iris_proba"] = 1.1157508543538652
    res["default_iris_sparse"] = 0.38
    res["default_digits"] = 0.81238615664845171
    res["digits_n_calls"] = None
    res["default_digits_iterative"] = 0.81238615664845171
    res["default_digits_binary"] = 0.99392835458409234
    res["default_digits_multilabel"] = 0.67651391068802913
    res["default_digits_multilabel_proba"] = 0.5

    sk_mod = sklearn.naive_bayes.BernoulliNB
    module = BernoulliNB



from typing import Optional

from functools import partial

from ConfigSpace.configuration_space import ConfigurationSpace
from ConfigSpace.hyperparameters import (
    CategoricalHyperparameter,
    UniformFloatHyperparameter,
)

from autosklearn.askl_typing import FEAT_TYPE_TYPE
from autosklearn.pipeline.components.base import AutoSklearnPreprocessingAlgorithm
from autosklearn.pipeline.components.feature_preprocessing.select_percentile import (
    SelectPercentileBase,
)
from autosklearn.pipeline.constants import DENSE, INPUT, SPARSE, UNSIGNED_DATA


class SelectPercentileRegression(
    SelectPercentileBase, AutoSklearnPreprocessingAlgorithm
):
    def __init__(self, percentile, score_func="f_regression", random_state=None):
        """Parameters:
        random state : ignored

        score_func : callable, Function taking two arrays X and y, and
                     returning a pair of arrays (scores, pvalues).
        """
        import sklearn.feature_selection

        self.random_state = random_state  # We don't use this
        self.percentile = int(float(percentile))
        if score_func == "f_regression":
            self.score_func = sklearn.feature_selection.f_regression
        elif score_func == "mutual_info":
            self.score_func = partial(
                sklearn.feature_selection.mutual_info_regression,
                random_state=self.random_state,
            )
        else:
            raise ValueError("Don't know this scoring function: %s" % score_func)

    @staticmethod
    def get_properties(dataset_properties=None):
        return {
            "shortname": "SPR",
            "name": "Select Percentile Regression",
            "handles_regression": True,
            "handles_classification": False,
            "handles_multiclass": False,
            "handles_multilabel": False,
            "handles_multioutput": False,
            "is_deterministic": True,
            "input": (DENSE, SPARSE, UNSIGNED_DATA),
            "output": (INPUT,),
        }

    @staticmethod
    def get_hyperparameter_search_space(
        feat_type: Optional[FEAT_TYPE_TYPE] = None, dataset_properties=None
    ):
        percentile = UniformFloatHyperparameter(
            "percentile", lower=1, upper=99, default_value=50
        )

        score_func = CategoricalHyperparameter(
            name="score_func", choices=["f_regression", "mutual_info"]
        )

        cs = ConfigurationSpace()
        cs.add_hyperparameters([percentile, score_func])
        return cs



from typing import Any, Dict

import sklearn.neural_network

from autosklearn.pipeline.components.regression.mlp import MLPRegressor

from .test_base import BaseRegressionComponentTest


class MLPComponentTest(BaseRegressionComponentTest):
    # NOTE: `default_boston`
    #
    # Github runners seem to indeterministicly fail `test_boston`
    # meaning 'default_irish_proba_places' needs to be set.
    # There are known differences to occur on different platforms.
    # https://github.com/scikit-learn/scikit-learn/issues/13108#issuecomment-461696681
    #
    # We are assuming results are deterministic on a given platform as locally
    # there is no randomness i.e. performing the same test 100 times yeilds the
    # same predictions 100 times.
    #
    # Github runners indicate that they run on microsoft Azure with DS2-v2.
    # https://docs.github.com/en/actions/using-github-hosted-runners/about-github-hosted-runners#cloud-hosts-for-github-hosted-runners
    #
    # These seem to have consistent CPU's so I'm unsure what the underlying reason
    # for this to randomly fail only sometimes on Github runners
    #
    # Edit: If changing, please tracke what values were failing
    #
    # Seems there is a consistently different values for boston so:
    # * include two valuess for n_iter in 'boston_iterative_n_iter'
    #   known-values = [236, 331, 327]
    #
    # * decreased places from 6 -> 5 in 'default_boston_{sparse,_iterative_sparse}'
    #   to check for for iterations and expanded the default places for checking
    #   know-values = [-0.10972947168054104, -0.10973142976866268]
    #
    # * decreased places from 3 -> 1 in 'default_boston_places'
    #   known-values = [0.29521793994422807, 0.2750079862455884]
    #
    # * Include two value for 'boston_n_calls'
    #   known-values = [8, 9]
    __test__ = True
    __test__ = True

    res: Dict[str, Any] = {}
    res["default_boston"] = 0.2750079862455884
    res["default_boston_places"] = 1
    res["boston_n_calls"] = [8, 9]
    res["boston_iterative_n_iter"] = [236, 331, 327]
    res["default_boston_iterative"] = res["default_boston"]
    res["default_boston_iterative_places"] = 1
    res["default_boston_sparse"] = -0.10972947168054104
    res["default_boston_sparse_places"] = 5
    res["default_boston_iterative_sparse"] = res["default_boston_sparse"]
    res["default_boston_iterative_sparse_places"] = res["default_boston_sparse_places"]
    res["default_diabetes"] = 0.35917389841850555
    res["diabetes_n_calls"] = 9
    res["diabetes_iterative_n_iter"] = 435
    res["default_diabetes_iterative"] = res["default_diabetes"]
    res["default_diabetes_sparse"] = 0.25573903970369427
    res["default_diabetes_iterative_sparse"] = res["default_diabetes_sparse"]

    sk_mod = sklearn.neural_network.MLPRegressor
    module = MLPRegressor
    step_hyperparameter = {
        "name": "n_iter_",
        "value": module.get_max_iter(),
    }



from typing import Container, Type

import numpy as np
import sklearn.metrics

from autosklearn.pipeline.components.regression import RegressorChoice, _regressors
from autosklearn.pipeline.components.regression.libsvm_svr import LibSVM_SVR
from autosklearn.pipeline.constants import SPARSE
from autosklearn.pipeline.util import _test_regressor, _test_regressor_iterative_fit

import pytest
import unittest

from test.test_pipeline.ignored_warnings import ignore_warnings, regressor_warnings


class BaseRegressionComponentTest(unittest.TestCase):

    res = None

    module = None
    sk_module = None
    # Hyperparameter which is increased by iterative_fit
    step_hyperparameter = None

    # Magic command to not run tests on base class
    __test__ = False

    def test_default_boston(self):

        if self.__class__ == BaseRegressionComponentTest:
            return

        for _ in range(2):

            with ignore_warnings(regressor_warnings):
                predictions, targets, n_calls = _test_regressor(
                    dataset="boston", Regressor=self.module
                )

            score = sklearn.metrics.r2_score(y_true=targets, y_pred=predictions)

            # Special treatment for Gaussian Process Regression
            if "default_boston_le_ge" in self.res:
                upper, lower = self.res["default_boston_le_ge"]
                assert lower <= score <= upper

            else:
                fixture = self.res["default_boston"]
                places = self.res.get("default_boston_places", 7)

                if score < -1e10:
                    score = np.log(-score)
                    fixture = np.log(-fixture)

                self.assertAlmostEqual(fixture, score, places)

            if "boston_n_calls" in self.res:
                expected = self.res["boston_n_calls"]
                if isinstance(expected, Container):
                    assert n_calls in expected
                else:
                    assert n_calls == expected

    def test_default_boston_iterative_fit(self):

        if self.__class__ == BaseRegressionComponentTest:
            return

        if not hasattr(self.module, "iterative_fit"):
            return

        for i in range(2):
            with ignore_warnings(regressor_warnings):
                predictions, targets, regressor = _test_regressor_iterative_fit(
                    dataset="boston", Regressor=self.module
                )

            score = sklearn.metrics.r2_score(targets, predictions)
            fixture = self.res["default_boston_iterative"]
            places = self.res.get("default_boston_iterative_places", 7)

            if score < -1e10:
                print(f"score = {score}, fixture = {fixture}")
                score = np.log(-score)
                fixture = np.log(-fixture)

            self.assertAlmostEqual(fixture, score, places)

            if self.step_hyperparameter is not None:
                param_name = self.step_hyperparameter["name"]
                default = self.step_hyperparameter["value"]

                value = getattr(regressor.estimator, param_name)
                expected = self.res.get("boston_iterative_n_iter", default)

                # To currently allow for MLPRegressor which is indeterministic,
                # we can have multiple values
                if isinstance(expected, Container):
                    assert value in expected
                else:
                    assert value == expected

    def test_default_boston_iterative_sparse_fit(self):

        if self.__class__ == BaseRegressionComponentTest:
            return

        if not hasattr(self.module, "iterative_fit"):
            return

        if SPARSE not in self.module.get_properties()["input"]:
            return

        for i in range(2):
            with ignore_warnings(regressor_warnings):
                predictions, targets, _ = _test_regressor_iterative_fit(
                    dataset="boston", Regressor=self.module, sparse=True
                )
            self.assertAlmostEqual(
                self.res["default_boston_iterative_sparse"],
                sklearn.metrics.r2_score(targets, predictions),
                places=self.res.get("default_boston_iterative_sparse_places", 7),
            )

    def test_default_boston_sparse(self):

        if self.__class__ == BaseRegressionComponentTest:
            return

        if SPARSE not in self.module.get_properties()["input"]:
            return

        for i in range(2):
            with ignore_warnings(regressor_warnings):
                predictions, targets, _ = _test_regressor(
                    dataset="boston", Regressor=self.module, sparse=True
                )

            self.assertAlmostEqual(
                self.res["default_boston_sparse"],
                sklearn.metrics.r2_score(targets, predictions),
                places=self.res.get("default_boston_sparse_places", 7),
            )

    def test_default_diabetes(self):

        if self.__class__ == BaseRegressionComponentTest:
            return

        for i in range(2):
            with ignore_warnings(regressor_warnings):
                predictions, targets, n_calls = _test_regressor(
                    dataset="diabetes", Regressor=self.module
                )

            self.assertAlmostEqual(
                self.res["default_diabetes"],
                sklearn.metrics.r2_score(targets, predictions),
                places=self.res.get("default_diabetes_places", 7),
            )

            if self.res.get("diabetes_n_calls"):
                self.assertEqual(self.res["diabetes_n_calls"], n_calls)

    def test_default_diabetes_iterative_fit(self):

        if self.__class__ == BaseRegressionComponentTest:
            return

        if not hasattr(self.module, "iterative_fit"):
            return

        for i in range(2):
            with ignore_warnings(regressor_warnings):
                predictions, targets, _ = _test_regressor_iterative_fit(
                    dataset="diabetes", Regressor=self.module
                )

            self.assertAlmostEqual(
                self.res["default_diabetes_iterative"],
                sklearn.metrics.r2_score(targets, predictions),
                places=self.res.get("default_diabetes_iterative_places", 7),
            )

    def test_default_diabetes_iterative_sparse_fit(self):

        if self.__class__ == BaseRegressionComponentTest:
            return

        if not hasattr(self.module, "iterative_fit"):
            return

        if SPARSE not in self.module.get_properties()["input"]:
            return

        for i in range(2):
            with ignore_warnings(regressor_warnings):
                predictions, targets, regressor = _test_regressor_iterative_fit(
                    dataset="diabetes", Regressor=self.module, sparse=True
                )

            self.assertAlmostEqual(
                self.res["default_diabetes_iterative_sparse"],
                sklearn.metrics.r2_score(targets, predictions),
                places=self.res.get("default_diabetes_iterative_sparse_places", 7),
            )

            if self.step_hyperparameter is not None:
                self.assertEqual(
                    getattr(regressor.estimator, self.step_hyperparameter["name"]),
                    self.res.get(
                        "diabetes_iterative_n_iter", self.step_hyperparameter["value"]
                    ),
                )

    def test_default_diabetes_sparse(self):

        if self.__class__ == BaseRegressionComponentTest:
            return

        if SPARSE not in self.module.get_properties()["input"]:
            return

        for i in range(2):
            with ignore_warnings(regressor_warnings):
                predictions, targets, _ = _test_regressor(
                    dataset="diabetes", Regressor=self.module, sparse=True
                )

            self.assertAlmostEqual(
                self.res["default_diabetes_sparse"],
                sklearn.metrics.r2_score(targets, predictions),
                places=self.res.get("default_diabetes_sparse_places", 7),
            )

    def test_module_idempotent(self):
        """Fitting twice with the same config gives the same model params.

        This is only valid when the random_state passed is an int. If a
        RandomState object is passed then repeated calls to fit will have
        different results. See the section on "Controlling Randomness" in the
        sklearn docs.

        https://scikit-learn.org/0.24/common_pitfalls.html#controlling-randomness
        """
        if self.__class__ == BaseRegressionComponentTest:
            return

        regressor_cls = self.module

        X = np.array(
            [
                [0.5, 0.5],
                [0.5, 0.5],
                [0.5, 0.5],
                [0.5, 0.5],
                [0.5, 0.5],
                [0.5, 0.5],
                [0.5, 0.5],
                [0.5, 0.5],
                [0.5, 0.5],
                [0.5, 0.5],
                [0.5, 0.5],
                [0.5, 0.5],
                [0.5, 0.5],
                [0.5, 0.5],
                [0.5, 0.5],
                [0.5, 0.5],
            ]
        )
        y = np.array(
            [
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
            ]
        )

        # We ignore certain keys when comparing
        param_keys_ignored = ["base_estimator"]

        # We use the default config + sampled ones
        configuration_space = regressor_cls.get_hyperparameter_search_space()

        default = configuration_space.get_default_configuration()
        sampled = [configuration_space.sample_configuration() for _ in range(2)]

        for seed, config in enumerate([default] + sampled):
            model_args = {"random_state": seed, **config}
            regressor = regressor_cls(**model_args)

            # Get the parameters on the first and second fit with config params
            # Also compare their random state
            with ignore_warnings(regressor_warnings):
                params_first = regressor.fit(X.copy(), y.copy()).estimator.get_params()

            if hasattr(regressor.estimator, "random_state"):
                rs_1 = regressor.random_state
                rs_estimator_1 = regressor.estimator.random_state

            with ignore_warnings(regressor_warnings):
                params_second = regressor.fit(X.copy(), y.copy()).estimator.get_params()

            if hasattr(regressor.estimator, "random_state"):
                rs_2 = regressor.random_state
                rs_estimator_2 = regressor.estimator.random_state

            # Remove keys we don't wish to include in the comparison
            for params in [params_first, params_second]:
                for key in param_keys_ignored:
                    if key in params:
                        del params[key]

            # They should have equal parameters
            self.assertEqual(
                params_first, params_second, f"Failed with model args {model_args}"
            )
            if hasattr(regressor.estimator, "random_state") and not isinstance(
                regressor, LibSVM_SVR
            ):
                # sklearn.svm.SVR has it as an attribute but does not use it and
                # defaults it to None, even if a value is passed in
                assert all(
                    [
                        seed == random_state
                        for random_state in [rs_1, rs_estimator_1, rs_2, rs_estimator_2]
                    ]
                )


@pytest.mark.parametrize("regressor", _regressors.values())
@pytest.mark.parametrize("X", [np.array([[1, 2, 3]] * 20)])
@pytest.mark.parametrize("y", [np.array([1] * 20)])
def test_fit_and_predict_with_1d_targets_as_1d(
    regressor: Type[RegressorChoice], X: np.ndarray, y: np.ndarray
) -> None:
    """Test that all pipelines work with 1d target types

    Parameters
    ----------
    regressor: RegressorChoice
        The regressor to test

    X: np.ndarray
        The features

    y: np.ndarray
        The 1d targets

    Expects
    -------
    * Should be able to fit with 1d targets
    * Should be able to predict with 1d targest
    * Should have predictions with the same shape as y
    """
    assert len(X) == len(y)
    assert y.ndim == 1

    config_space = regressor.get_hyperparameter_search_space()
    default_config = config_space.get_default_configuration()

    model = regressor(random_state=0, **default_config)

    with ignore_warnings(regressor_warnings):
        model.fit(X, y)

    predictions = model.predict(X)

    assert predictions.shape == y.shape


@pytest.mark.parametrize("regressor", _regressors.values())
@pytest.mark.parametrize("X", [np.array([[1, 2, 3]] * 20)])
@pytest.mark.parametrize("y", [np.array([[1]] * 20)])
def test_fit_and_predict_with_1d_targets_as_2d(
    regressor: Type[RegressorChoice], X: np.ndarray, y: np.ndarray
) -> None:
    """Test that all pipelines work with 1d target types when they are wrapped as 2d

    Parameters
    ----------
    regressor: RegressorChoice
        The regressor to test

    X: np.ndarray
        The features

    y: np.ndarray
        The 1d targets wrapped as 2d

    Expects
    -------
    * Should be able to fit with 1d targets wrapped in 2d
    * Should be able to predict 1d targets wrapped in 2d
    * Should return 1d predictions
    * Should have predictions with the same length as the y
    """
    assert len(X) == len(y)
    assert y.ndim == 2 and y.shape[1] == 1

    config_space = regressor.get_hyperparameter_search_space()
    default_config = config_space.get_default_configuration()

    model = regressor(random_state=0, **default_config)

    with ignore_warnings(regressor_warnings):
        model.fit(X, y)

    predictions = model.predict(X)

    assert predictions.ndim == 1
    assert len(predictions) == len(y)


@pytest.mark.parametrize(
    "regressor",
    [
        regressor
        for regressor in _regressors.values()
        if regressor.get_properties()["handles_multilabel"]
    ],
)
@pytest.mark.parametrize("X", [np.array([[1, 2, 3]] * 20)])
@pytest.mark.parametrize("y", [np.array([[1, 1, 1]] * 20)])
def test_fit_and_predict_with_2d_targets(
    regressor: Type[RegressorChoice], X: np.ndarray, y: np.ndarray
) -> None:
    """Test that all pipelines work with 2d target types

    Parameters
    ----------
    regressor: RegressorChoice
        The regressor to test

    X: np.ndarray
        The features

    y: np.ndarray
        The 2d targets

    Expects
    -------
    * Should be able to fit with 2d targets
    * Should be able to predict with 2d targets
    * Should have predictions with the same shape as y
    """
    assert len(X) == len(y)
    assert y.ndim == 2 and y.shape[1] > 1

    config_space = regressor.get_hyperparameter_search_space()
    default_config = config_space.get_default_configuration()

    model = regressor(random_state=0, **default_config)

    with ignore_warnings(regressor_warnings):
        model.fit(X, y)

    predictions = model.predict(X)

    assert predictions.shape == y.shape



"""Test that autosklearn is sklearn compliant

Note
----
* This is far from complete at the moment
* This should probably be tested on AutoSklearnEstimators not AutoML
"""
from typing import List, Union

from sklearn.exceptions import NotFittedError

from autosklearn.automl import AutoML

import pytest
from pytest_cases import parametrize, parametrize_with_cases
from pytest_cases.filters import has_tag

import test.test_automl.cases as cases


@pytest.mark.xfail(
    reason="__sklearn_is_fitted__ only supported from sklearn 1.0 onwards"
)
@parametrize_with_cases("automl", cases=cases, filter=~has_tag("fitted"))
@parametrize(
    "attr, argnames",
    [
        ("refit", ["X", "y"]),
        ("predict", ["X"]),
        ("fit_ensemble", ["y"]),
        ("score", ["X", "y"]),
        ("performance_over_time_", None),
        ("cv_results_", None),
        ("sprint_statistics", []),
        ("get_models_with_weights", []),
        ("show_models", []),
    ],
)
def test_attrs_raise_if_not_fitted(
    automl: AutoML,
    attr: str,
    argnames: Union[List[str], None],
) -> None:
    """
    Parameters
    ----------
    automl : AutoML
        An unfitted automl instance

    attr: str
        The attribute to test

    argnames: Union[List[str], None]
        The arguments of the the method
        * ["arg1", "arg2"] for method with args
        * [] for method with no args
        * None for property

    Expects
    -------
    * Should raise a NotFittedError

    Note
    ----
    * This also ensures any validation should be done after the fit check as
    NotFittedError should be raised
    """
    with pytest.raises(NotFittedError):

        if argnames is None:
            property = getattr(automl, attr)  # noqa
        else:
            method = getattr(automl, attr)
            args = {name: None for name in argnames}
            if len(args) > 0:
                method(args)
            else:
                method()



import sklearn.metrics
from sklearn.linear_model import Ridge

from autosklearn.pipeline.components.feature_preprocessing.fast_ica import FastICA
from autosklearn.pipeline.util import (
    PreprocessingTestCase,
    _test_preprocessing,
    get_dataset,
)

import unittest


class FastICAComponentTest(PreprocessingTestCase):
    def test_default_configuration(self):
        transformation, original = _test_preprocessing(FastICA, dataset="diabetes")
        self.assertEqual(transformation.shape[0], original.shape[0])
        self.assertFalse((transformation == 0).all())

    def test_default_configuration_regression(self):
        for i in range(5):
            X_train, Y_train, X_test, Y_test = get_dataset(dataset="diabetes")
            configuration_space = FastICA.get_hyperparameter_search_space()
            default = configuration_space.get_default_configuration()
            preprocessor = FastICA(
                random_state=1, **{hp_name: default[hp_name] for hp_name in default}
            )
            preprocessor.fit(X_train, Y_train)
            X_train_trans = preprocessor.transform(X_train)
            X_test_trans = preprocessor.transform(X_test)

            # fit a classifier on top
            classifier = Ridge()
            predictor = classifier.fit(X_train_trans, Y_train)
            predictions = predictor.predict(X_test_trans)
            accuracy = sklearn.metrics.r2_score(Y_test, predictions)
            self.assertAlmostEqual(accuracy, 0.32614416980439365)

    @unittest.skip("Always returns float64")
    def test_preprocessing_dtype(self):
        super(FastICAComponentTest, self)._test_preprocessing_dtype(
            FastICA, dataset="diabetes"
        )



"""Test the performance of automl instances after fitting"""

import numpy as np
from sklearn.ensemble import VotingClassifier, VotingRegressor

from autosklearn.automl import AutoML

from pytest_cases import parametrize_with_cases

import test.test_automl.cases as cases


@parametrize_with_cases("automl", cases.case_classifier_fitted_holdout_multiobjective)
def test_performance_with_multiobjective(automl: AutoML) -> None:
    """
    Expects
    -------
    * Auto-sklearn can predict/predict_proba and has a model
    * Each ensemble in the pareto_set can predict/predict_proba
    """
    # TODO: This test is hyperspecific to this one case
    #
    #   Long term we probably want to return additional info about the case so we can
    #   test things for other than this case

    # Check that the predict function works
    X = np.array([[1.0, 1.0, 1.0, 1.0]])

    assert automl.predict_proba(X).shape == (1, 3)
    assert automl.predict(X).shape == (1,)

    pareto_front = automl._load_pareto_set()
    for ensemble in pareto_front:

        assert isinstance(ensemble, (VotingClassifier, VotingRegressor))

        y_pred = ensemble.predict_proba(X)
        assert y_pred.shape == (1, 3)

        y_pred = ensemble.predict(X)
        assert y_pred in ["setosa", "versicolor", "virginica"]

    statistics = automl.sprint_statistics()
    assert "Metrics" in statistics
    assert ("Best validation score: 0.9" in statistics) or (
        "Best validation score: 1.0" in statistics
    ), statistics



import numpy as np
import pandas as pd
from scipy import sparse

from autosklearn.pipeline.components.data_preprocessing.imputation.categorical_imputation import (  # noqa: E501
    CategoricalImputation,
)

import pytest


@pytest.fixture
def input_data_imputation(request):
    size = (50, 20)
    X = np.array(np.random.randint(3, 10, size=size), dtype=float)
    mask = np.logical_not(np.random.randint(0, 5, size=size), dtype=bool)
    X[mask] = np.nan
    if request.param == "numpy":
        pass
    elif request.param == "pandas":
        X = pd.DataFrame(X)
    return X, mask


@pytest.mark.parametrize("input_data_imputation", ("numpy", "pandas"), indirect=True)
@pytest.mark.parametrize("categorical", (True, False))
def test_default_imputation(input_data_imputation, categorical):
    """
    Makes sure that imputation works for both numerical and categorical data.
    This also has to be guaranteed for numpy and pandas like objects.
    """
    X, mask = input_data_imputation
    if categorical:
        imputation_value = "missing_value"
        X = X.astype("str").astype("object")
        X[mask] = np.nan
    else:
        imputation_value = min(np.unique(X)) - 1

    Y = CategoricalImputation().fit_transform(X.copy())

    assert np.array_equal(Y == imputation_value, mask)
    assert np.array_equal(Y != imputation_value, ~mask)


@pytest.mark.parametrize("format_type", ("numpy", "pandas"))
def test_nonzero_numerical_imputation(format_type):
    # First try with an array with 0 as only valid category. The imputation should
    # happen with -1
    X = np.full(fill_value=np.nan, shape=(10, 10))
    X[0, :] = 0
    if "pandas" in format_type:
        X = pd.DataFrame(X)
    elif "numpy" in format_type:
        pass
    else:
        pytest.fail(format_type)
    Y = CategoricalImputation().fit_transform(X.copy())
    np.testing.assert_equal(np.nan_to_num(X, nan=-1, copy=True), Y)

    # Then if there is also a -1 in the category, we expect -2 as imputation
    X = np.full(fill_value=np.nan, shape=(10, 10))
    X[0, :] = 0
    X[1, :] = -1
    if "pandas" in format_type:
        X = pd.DataFrame(X)
    Y = CategoricalImputation().fit_transform(X.copy())
    np.testing.assert_equal(np.nan_to_num(X, nan=-2, copy=True), Y)


@pytest.mark.parametrize("input_data_imputation", ("numpy"), indirect=True)
def test_default_sparse(input_data_imputation):
    X, mask = input_data_imputation
    X = sparse.csr_matrix(X)
    Y = CategoricalImputation().fit_transform(X)
    Y = Y.todense()

    np.testing.assert_equal(Y == 0, mask)
    np.testing.assert_equal(Y != 0, ~mask)



import os
import sys

from setuptools import find_packages, setup

HERE = os.path.abspath(os.path.dirname(__file__))

# Check if Auto-sklearn *could* run on the given system
if os.name != "posix":
    raise ValueError(
        "Detected unsupported operating system: %s. Please check "
        "the compability information of auto-sklearn: https://automl.github.io"
        "/auto-sklearn/master/installation.html#windows-osx-compatibility"
        % sys.platform
    )

if sys.version_info < (3, 7):
    raise ValueError(
        "Unsupported Python version %d.%d.%d found. Auto-sklearn requires Python "
        "3.7 or higher."
        % (sys.version_info.major, sys.version_info.minor, sys.version_info.micro)
    )

with open(os.path.join(HERE, "requirements.txt")) as fp:
    install_reqs = [
        r.rstrip()
        for r in fp.readlines()
        if not r.startswith("#") and not r.startswith("git+")
    ]

extras_reqs = {
    "test": [
        "pytest>=4.6",
        "pytest-cov",
        "pytest-xdist",
        "pytest-timeout",
        "pytest-cases>=3.6.11",
        "mypy",
        "isort",
        "black",
        "pydocstyle",
        "openml",
        "pre-commit",
    ],
    "examples": [
        "matplotlib",
        "jupyter",
        "notebook",
        "seaborn",
    ],
    "docs": [
        "sphinx<4.3",
        "sphinx-gallery",
        "sphinx_bootstrap_theme",
        "numpydoc",
        "sphinx_toolbox",
        "docutils==0.16",
    ],
}

with open(os.path.join(HERE, "autosklearn", "__version__.py")) as fh:
    version = fh.readlines()[-1].split()[-1].strip("\"'")


with open(os.path.join(HERE, "README.md")) as fh:
    long_description = fh.read()


setup(
    name="auto-sklearn",
    author="Matthias Feurer",
    author_email="feurerm@informatik.uni-freiburg.de",
    description="Automated machine learning.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    version=version,
    packages=find_packages(exclude=["test", "scripts", "examples"]),
    extras_require=extras_reqs,
    install_requires=install_reqs,
    include_package_data=True,
    license="BSD3",
    platforms=["Linux"],
    classifiers=[
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Information Technology",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    python_requires=">=3.7",
    url="https://automl.github.io/auto-sklearn",
)



import numpy as np

from autosklearn.constants import BINARY_CLASSIFICATION, REGRESSION
from autosklearn.ensembles.ensemble_selection import EnsembleSelection
from autosklearn.metrics import accuracy, root_mean_squared_error

import pytest


def testEnsembleSelection(backend):
    """
    Makes sure ensemble selection fit method creates an ensemble correctly
    """

    ensemble = EnsembleSelection(
        ensemble_size=10,
        task_type=REGRESSION,
        random_state=0,
        backend=backend,
        metrics=[root_mean_squared_error],
    )

    # We create a problem such that we encourage the addition of members to the ensemble
    # Fundamentally, the average of 10 sequential number is 5.5
    # X_data will be ignored and is therefore random
    X_data = np.random.random(size=(100, 2))
    y_true = np.full((100), 5.5)
    predictions = []
    for i in range(1, 20):
        pred = np.full((100), i, dtype=np.float32)
        pred[i * 5 : 5 * (i + 1)] = 5.5 * i
        predictions.append(pred)

    ensemble.fit(
        base_models_predictions=predictions,
        X_data=X_data,
        true_targets=y_true,
        model_identifiers=[(i, i, i) for i in range(20)],
        runs=[],
    )

    np.testing.assert_array_equal(
        ensemble.weights_,
        np.array(
            [
                0.1,
                0.2,
                0.2,
                0.1,
                0.1,
                0.1,
                0.1,
                0.1,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
            ]
        ),
    )

    assert ensemble.identifiers_ == [(i, i, i) for i in range(20)]

    np.testing.assert_array_almost_equal(
        np.array(ensemble.trajectory_),
        np.array(
            [
                3.462296925452813,
                2.679202306657711,
                2.2748626436960375,
                2.065717187806695,
                1.7874562615598728,
                1.6983448128441783,
                1.559451106330085,
                1.5316326052614575,
                1.3801950121782542,
                1.3554980575295374,
            ]
        ),
    )


def testPredict(backend):
    # Test that ensemble prediction applies weights correctly to given
    # predictions. There are two possible cases:
    # 1) predictions.shape[0] == len(self.weights_). In this case,
    # predictions include those made by zero-weighted models. Therefore,
    # we simply apply each weights to the corresponding model preds.
    # 2) predictions.shape[0] < len(self.weights_). In this case,
    # predictions exclude those made by zero-weighted models. Therefore,
    # we first exclude all occurrences of zero in self.weights_, and then
    # apply the weights.
    # If none of the above is the case, predict() raises Error.
    ensemble = EnsembleSelection(
        ensemble_size=3,
        task_type=BINARY_CLASSIFICATION,
        random_state=0,
        backend=backend,
        metrics=[accuracy],
    )
    # Test for case 1. Create (3, 2, 2) predictions.
    per_model_pred = np.array(
        [[[0.9, 0.1], [0.4, 0.6]], [[0.8, 0.2], [0.3, 0.7]], [[1.0, 0.0], [0.1, 0.9]]]
    )
    # Weights of 3 hypothetical models
    ensemble.weights_ = [0.7, 0.2, 0.1]
    pred = ensemble.predict(per_model_pred)
    truth = np.array(
        [[0.89, 0.11], [0.35, 0.65]]  # This should be the true prediction.
    )
    assert np.allclose(pred, truth)

    # Test for case 2.
    per_model_pred = np.array(
        [[[0.9, 0.1], [0.4, 0.6]], [[0.8, 0.2], [0.3, 0.7]], [[1.0, 0.0], [0.1, 0.9]]]
    )
    # The third model now has weight of zero.
    ensemble.weights_ = [0.7, 0.2, 0.0, 0.1]
    pred = ensemble.predict(per_model_pred)
    truth = np.array([[0.89, 0.11], [0.35, 0.65]])
    assert np.allclose(pred, truth)

    # Test for error case.
    per_model_pred = np.array(
        [[[0.9, 0.1], [0.4, 0.6]], [[0.8, 0.2], [0.3, 0.7]], [[1.0, 0.0], [0.1, 0.9]]]
    )
    # Now the weights have 2 zero weights and 2 non-zero weights,
    # which is incompatible.
    ensemble.weights_ = [0.6, 0.0, 0.0, 0.4]

    with pytest.raises(ValueError):
        ensemble.predict(per_model_pred)



import logging
import os
import tempfile

import arff
import numpy as np
import pandas as pd
from joblib import Memory
from sklearn.datasets import fetch_openml, make_multilabel_classification

import autosklearn.metalearning.metafeatures.metafeatures as meta_features
from autosklearn.metalearning.metafeatures.metafeature import MetaFeatureValue
from autosklearn.pipeline.components.data_preprocessing.feature_type import (
    FeatTypeSplit,
)

import pytest
import unittest


@pytest.fixture(scope="class", params=("pandas", "numpy"))
def multilabel_train_data(request):
    cache = Memory(location=tempfile.gettempdir())
    cached_func = cache.cache(make_multilabel_classification)
    X, y = cached_func(
        n_samples=100,
        n_features=10,
        n_classes=5,
        n_labels=5,
        return_indicator=True,
        random_state=1,
    )
    if request.param == "numpy":
        return X, y
    elif request.param == "pandas":
        return pd.DataFrame(X), y
    else:
        raise ValueError(request.param)


@pytest.fixture(scope="class", params=("pandas", "numpy"))
def meta_train_data(request):
    tests_dir = __file__
    os.chdir(os.path.dirname(tests_dir))

    decoder = arff.ArffDecoder()
    with open(os.path.join("datasets", "dataset.arff")) as fh:
        dataset = decoder.decode(fh, encode_nominal=True)

    # -1 because the last attribute is the class
    feat_type = {
        idx: "numerical" if type(type_) != list else "categorical"
        for idx, (name, type_) in enumerate(dataset["attributes"][:-1])
    }

    data = np.array(dataset["data"], dtype=np.float64)
    X = data[:, :-1]
    y = data[:, -1].reshape((-1,))

    logger = logging.getLogger("Meta")
    meta_features.helper_functions.set_value(
        "MissingValues",
        meta_features.helper_functions["MissingValues"](X, y, logger, feat_type),
    )
    meta_features.helper_functions.set_value(
        "NumSymbols",
        meta_features.helper_functions["NumSymbols"](X, y, logger, feat_type),
    )
    meta_features.helper_functions.set_value(
        "ClassOccurences",
        meta_features.helper_functions["ClassOccurences"](X, y, logger),
    )
    if request.param == "numpy":
        return X, y, feat_type
    elif request.param == "pandas":
        dtypes = {}
        for key, value in feat_type.items():
            if value == "categorical":
                dtypes[key] = "category"
            elif value == "numerical":
                dtypes[key] = "float64"
            elif value == "string":
                dtypes[key] = "string"
            else:
                raise KeyError

        X = pd.DataFrame(X).astype(dtypes)
        return X, y, feat_type
    else:
        raise ValueError(request.param)


@pytest.fixture(scope="class", params=("pandas", "numpy"))
def meta_train_data_transformed(request):
    tests_dir = __file__
    os.chdir(os.path.dirname(tests_dir))

    decoder = arff.ArffDecoder()
    with open(os.path.join("datasets", "dataset.arff")) as fh:
        dataset = decoder.decode(fh, encode_nominal=True)

    # -1 because the last attribute is the class
    feat_type = {
        idx: "numerical" if type(type_) != list else "categorical"
        for idx, (name, type_) in enumerate(dataset["attributes"][:-1])
    }

    data = np.array(dataset["data"], dtype=np.float64)
    X = data[:, :-1]
    y = data[:, -1].reshape((-1,))

    logger = logging.getLogger("Meta")
    meta_features.helper_functions.set_value(
        "MissingValues",
        meta_features.helper_functions["MissingValues"](X, y, logger, feat_type),
    )
    meta_features.helper_functions.set_value(
        "NumSymbols",
        meta_features.helper_functions["NumSymbols"](X, y, logger, feat_type),
    )
    meta_features.helper_functions.set_value(
        "ClassOccurences",
        meta_features.helper_functions["ClassOccurences"](X, y, logger),
    )

    DPP = FeatTypeSplit(feat_type=feat_type)
    X_transformed = DPP.fit_transform(X)
    feat_type_transformed = {i: "numerical" for i in range(X_transformed.shape[1])}

    # pre-compute values for transformed inputs
    meta_features.helper_functions.set_value(
        "PCA",
        meta_features.helper_functions["PCA"](X_transformed, y, logger),
    )
    meta_features.helper_functions.set_value(
        "Skewnesses",
        meta_features.helper_functions["Skewnesses"](
            X_transformed, y, logger, feat_type_transformed
        ),
    )
    meta_features.helper_functions.set_value(
        "Kurtosisses",
        meta_features.helper_functions["Kurtosisses"](
            X_transformed, y, logger, feat_type_transformed
        ),
    )

    if request.param == "numpy":
        return X_transformed, y, feat_type_transformed
    elif request.param == "pandas":
        X_transformed = pd.DataFrame(X_transformed)
        return X_transformed, y, feat_type_transformed
    else:
        raise ValueError(request.param)


def test_number_of_instance(meta_train_data):
    X, y, feat_type = meta_train_data
    mf = meta_features.metafeatures["NumberOfInstances"](
        X, y, logging.getLogger("Meta"), feat_type
    )
    assert mf.value == 898
    assert isinstance(mf, MetaFeatureValue)


def test_number_of_classes(meta_train_data):
    X, y, feat_type = meta_train_data
    mf = meta_features.metafeatures["NumberOfClasses"](
        X, y, logging.getLogger("Meta"), feat_type
    )
    assert mf.value == 5
    assert isinstance(mf, MetaFeatureValue)


def test_number_of_features(meta_train_data):
    X, y, feat_type = meta_train_data
    mf = meta_features.metafeatures["NumberOfFeatures"](
        X, y, logging.getLogger("Meta"), feat_type
    )
    assert mf.value == 38
    assert isinstance(mf, MetaFeatureValue)


def test_missing_values(meta_train_data):
    X, y, feat_type = meta_train_data
    mf = meta_features.helper_functions["MissingValues"](
        X, y, logging.getLogger("Meta"), feat_type
    )
    assert isinstance(mf.value, pd.DataFrame if hasattr(X, "iloc") else np.ndarray)
    assert mf.value.shape == X.shape
    assert 22175 == np.count_nonzero(mf.value)


def test_number_of_Instances_with_missing_values(meta_train_data):
    X, y, feat_type = meta_train_data
    mf = meta_features.metafeatures["NumberOfInstancesWithMissingValues"](
        X, y, logging.getLogger("Meta"), feat_type
    )
    assert mf.value == 898
    assert isinstance(mf, MetaFeatureValue)


def test_percentage_of_Instances_with_missing_values(meta_train_data):
    X, y, feat_type = meta_train_data
    meta_features.metafeatures.set_value(
        "NumberOfInstancesWithMissingValues",
        meta_features.metafeatures["NumberOfInstancesWithMissingValues"](
            X, y, logging.getLogger("Meta"), feat_type
        ),
    )
    mf = meta_features.metafeatures["PercentageOfInstancesWithMissingValues"](
        X, y, logging.getLogger("Meta"), feat_type
    )
    assert pytest.approx(mf.value) == 1.0
    assert isinstance(mf, MetaFeatureValue)


def test_number_of_features_with_missing_values(meta_train_data):
    X, y, feat_type = meta_train_data
    mf = meta_features.metafeatures["NumberOfFeaturesWithMissingValues"](
        X, y, logging.getLogger("Meta"), feat_type
    )
    assert mf.value == 29
    assert isinstance(mf, MetaFeatureValue)


def test_percentage_of_features_with_missing_values(meta_train_data):
    X, y, feat_type = meta_train_data
    meta_features.metafeatures.set_value(
        "NumberOfFeaturesWithMissingValues",
        meta_features.metafeatures["NumberOfFeaturesWithMissingValues"](
            X, y, logging.getLogger("Meta"), feat_type
        ),
    )
    mf = meta_features.metafeatures["PercentageOfFeaturesWithMissingValues"](
        X, y, logging.getLogger("Meta"), feat_type
    )
    assert pytest.approx(mf.value) == float(29) / float(38)
    assert isinstance(mf, MetaFeatureValue)


def test_number_of_missing_values(meta_train_data):
    X, y, feat_type = meta_train_data
    np.save("/tmp/debug", X)
    mf = meta_features.metafeatures["NumberOfMissingValues"](
        X, y, logging.getLogger("Meta"), feat_type
    )
    assert mf.value == 22175
    assert isinstance(mf, MetaFeatureValue)


def test_percentage_missing_values(meta_train_data):
    X, y, feat_type = meta_train_data
    meta_features.metafeatures.set_value(
        "NumberOfMissingValues",
        meta_features.metafeatures["NumberOfMissingValues"](
            X, y, logging.getLogger("Meta"), feat_type
        ),
    )
    mf = meta_features.metafeatures["PercentageOfMissingValues"](
        X, y, logging.getLogger("Meta"), feat_type
    )
    assert pytest.approx(mf.value) == (float(22175) / float(38 * 898))
    assert isinstance(mf, MetaFeatureValue)


def test_number_of_numeric_features(meta_train_data):
    X, y, feat_type = meta_train_data
    mf = meta_features.metafeatures["NumberOfNumericFeatures"](
        X, y, logging.getLogger("Meta"), feat_type
    )
    assert mf.value == 6
    assert isinstance(mf, MetaFeatureValue)


def test_number_of_categorical_features(meta_train_data):
    X, y, feat_type = meta_train_data
    mf = meta_features.metafeatures["NumberOfCategoricalFeatures"](
        X, y, logging.getLogger("Meta"), feat_type
    )
    assert mf.value == 32
    assert isinstance(mf, MetaFeatureValue)


def test_ratio_numerical_to_categorical(meta_train_data):
    X, y, feat_type = meta_train_data
    mf = meta_features.metafeatures["RatioNumericalToNominal"](
        X, y, logging.getLogger("Meta"), feat_type
    )
    assert pytest.approx(mf.value) == (float(6) / float(32))
    assert isinstance(mf, MetaFeatureValue)


def test_ratio_categorical_to_numerical(meta_train_data):
    X, y, feat_type = meta_train_data
    mf = meta_features.metafeatures["RatioNominalToNumerical"](
        X, y, logging.getLogger("Meta"), feat_type
    )
    assert pytest.approx(mf.value) == (float(32) / float(6))
    assert isinstance(mf, MetaFeatureValue)


def test_dataset_ratio(meta_train_data):
    X, y, feat_type = meta_train_data
    mf = meta_features.metafeatures["DatasetRatio"](
        X, y, logging.getLogger("Meta"), feat_type
    )
    assert pytest.approx(mf.value) == (float(38) / float(898))
    assert isinstance(mf, MetaFeatureValue)


def test_inverse_dataset_ratio(meta_train_data):
    X, y, feat_type = meta_train_data
    mf = meta_features.metafeatures["InverseDatasetRatio"](
        X, y, logging.getLogger("Meta"), feat_type
    )
    assert pytest.approx(mf.value) == (float(898) / float(38))
    assert isinstance(mf, MetaFeatureValue)


def test_class_occurences(meta_train_data):
    X, y, feat_type = meta_train_data
    mf = meta_features.helper_functions["ClassOccurences"](
        X, y, logging.getLogger("Meta"), feat_type
    )
    assert mf.value == {0.0: 8.0, 1.0: 99.0, 2.0: 684.0, 4.0: 67.0, 5.0: 40.0}


def test_class_probability_min(meta_train_data):
    X, y, feat_type = meta_train_data
    mf = meta_features.metafeatures["ClassProbabilityMin"](
        X, y, logging.getLogger("Meta"), feat_type
    )
    assert pytest.approx(mf.value) == (float(8) / float(898))
    assert isinstance(mf, MetaFeatureValue)


def test_class_probability_max(meta_train_data):
    X, y, feat_type = meta_train_data
    mf = meta_features.metafeatures["ClassProbabilityMax"](
        X, y, logging.getLogger("Meta"), feat_type
    )
    assert pytest.approx(mf.value) == (float(684) / float(898))
    assert isinstance(mf, MetaFeatureValue)


def test_class_probability_mean(meta_train_data):
    X, y, feat_type = meta_train_data
    mf = meta_features.metafeatures["ClassProbabilityMean"](
        X, y, logging.getLogger("Meta"), feat_type
    )
    classes = np.array((8, 99, 684, 67, 40), dtype=np.float64)
    prob_mean = (classes / float(898)).mean()
    assert pytest.approx(mf.value) == prob_mean
    assert isinstance(mf, MetaFeatureValue)


def test_class_probability_std(meta_train_data):
    X, y, feat_type = meta_train_data
    mf = meta_features.metafeatures["ClassProbabilitySTD"](
        X, y, logging.getLogger("Meta"), feat_type
    )
    classes = np.array((8, 99, 684, 67, 40), dtype=np.float64)
    prob_std = (classes / float(898)).std()
    assert pytest.approx(mf.value) == prob_std
    assert isinstance(mf, MetaFeatureValue)


def test_num_symbols(meta_train_data):
    X, y, feat_type = meta_train_data
    mf = meta_features.helper_functions["NumSymbols"](
        X, y, logging.getLogger("Meta"), feat_type
    )
    symbol_frequency = [
        2,
        1,
        7,
        1,
        2,
        4,
        1,
        1,
        4,
        2,
        1,
        1,
        1,
        2,
        1,
        0,
        1,
        1,
        1,
        0,
        1,
        1,
        0,
        3,
        1,
        0,
        0,
        0,
        2,
        2,
        3,
        2,
    ]
    assert mf.value == symbol_frequency


def test_symbols_min(meta_train_data):
    X, y, feat_type = meta_train_data
    mf = meta_features.metafeatures["SymbolsMin"](
        X, y, logging.getLogger("Meta"), feat_type
    )
    assert mf.value == 1


def test_symbols_max(meta_train_data):
    X, y, feat_type = meta_train_data
    # this is attribute steel
    mf = meta_features.metafeatures["SymbolsMax"](
        X, y, logging.getLogger("Meta"), feat_type
    )
    assert mf.value == 7


def test_symbols_mean(meta_train_data):
    X, y, feat_type = meta_train_data
    mf = meta_features.metafeatures["SymbolsMean"](
        X, y, logging.getLogger("Meta"), feat_type
    )
    # Empty looking spaces denote empty attributes
    symbol_frequency = [
        2,
        1,
        7,
        1,
        2,
        4,
        1,
        1,
        4,
        2,
        1,
        1,
        1,
        2,
        1,  #
        1,
        1,
        1,
        1,
        1,
        3,
        1,
        2,
        2,
        3,
        2,
    ]
    assert pytest.approx(mf.value) == np.mean(symbol_frequency)


def test_symbols_std(meta_train_data):
    X, y, feat_type = meta_train_data
    mf = meta_features.metafeatures["SymbolsSTD"](
        X, y, logging.getLogger("Meta"), feat_type
    )
    symbol_frequency = [
        2,
        1,
        7,
        1,
        2,
        4,
        1,
        1,
        4,
        2,
        1,
        1,
        1,
        2,
        1,  #
        1,
        1,
        1,
        1,
        1,
        3,
        1,
        2,
        2,
        3,
        2,
    ]
    assert pytest.approx(mf.value) == np.std(symbol_frequency)


def test_symbols_sum(meta_train_data):
    X, y, feat_type = meta_train_data
    mf = meta_features.metafeatures["SymbolsSum"](
        X, y, logging.getLogger("Meta"), feat_type
    )
    assert mf.value == 49


def test_class_entropy(meta_train_data):
    X, y, feat_type = meta_train_data
    mf = meta_features.metafeatures["ClassEntropy"](
        X, y, logging.getLogger("Meta"), feat_type
    )
    classes = np.array((8, 99, 684, 67, 40), dtype=np.float64)
    classes = classes / sum(classes)
    entropy = -np.sum([c * np.log2(c) for c in classes])

    assert pytest.approx(mf.value) == entropy


def test_calculate_all_metafeatures(meta_train_data):
    X, y, feat_type = meta_train_data
    mf = meta_features.calculate_all_metafeatures(
        X, y, feat_type, "2", logger=logging.getLogger("Meta")
    )
    assert 52 == len(mf.metafeature_values)
    assert mf.metafeature_values["NumberOfCategoricalFeatures"].value == 32


def test_kurtosisses(meta_train_data_transformed):
    X_transformed, y, feat_type_transformed = meta_train_data_transformed
    mf = meta_features.helper_functions["Kurtosisses"](
        X_transformed, y, logging.getLogger("Meta"), feat_type_transformed
    )
    assert 81 == len(mf.value)


def test_kurtosis_min(meta_train_data_transformed):
    X_transformed, y, feat_type_transformed = meta_train_data_transformed
    # TODO: somehow compute the expected output?
    meta_features.metafeatures["KurtosisMin"](
        X_transformed, y, logging.getLogger("Meta"), feat_type_transformed
    )


def test_kurtosis_max(meta_train_data_transformed):
    X_transformed, y, feat_type_transformed = meta_train_data_transformed
    # TODO: somehow compute the expected output?
    meta_features.metafeatures["KurtosisMax"](
        X_transformed, y, logging.getLogger("Meta"), feat_type_transformed
    )


def test_kurtosis_mean(meta_train_data_transformed):
    X_transformed, y, feat_type_transformed = meta_train_data_transformed
    # TODO: somehow compute the expected output?
    meta_features.metafeatures["KurtosisMean"](
        X_transformed, y, logging.getLogger("Meta"), feat_type_transformed
    )


def test_kurtosis_std(meta_train_data_transformed):
    X_transformed, y, feat_type_transformed = meta_train_data_transformed
    # TODO: somehow compute the expected output?
    meta_features.metafeatures["KurtosisSTD"](
        X_transformed, y, logging.getLogger("Meta"), feat_type_transformed
    )


def test_skewnesses(meta_train_data_transformed):
    X_transformed, y, feat_type_transformed = meta_train_data_transformed
    mf = meta_features.helper_functions["Skewnesses"](
        X_transformed, y, logging.getLogger("Meta"), feat_type_transformed
    )
    assert 81 == len(mf.value)


def test_skewness_min(meta_train_data_transformed):
    X_transformed, y, feat_type_transformed = meta_train_data_transformed
    # TODO: somehow compute the expected output?
    meta_features.metafeatures["SkewnessMin"](
        X_transformed, y, logging.getLogger("Meta"), feat_type_transformed
    )


def test_skewness_max(meta_train_data_transformed):
    X_transformed, y, feat_type_transformed = meta_train_data_transformed
    # TODO: somehow compute the expected output?
    meta_features.metafeatures["SkewnessMax"](
        X_transformed, y, logging.getLogger("Meta"), feat_type_transformed
    )


def test_skewness_mean(meta_train_data_transformed):
    X_transformed, y, feat_type_transformed = meta_train_data_transformed
    # TODO: somehow compute the expected output?
    meta_features.metafeatures["SkewnessMean"](
        X_transformed, y, logging.getLogger("Meta"), feat_type_transformed
    )


def test_skewness_std(meta_train_data_transformed):
    X_transformed, y, feat_type_transformed = meta_train_data_transformed
    # TODO: somehow compute the expected output?
    meta_features.metafeatures["SkewnessSTD"](
        X_transformed, y, logging.getLogger("Meta"), feat_type_transformed
    )


def test_landmark_lda(meta_train_data_transformed):
    X_transformed, y, feat_type_transformed = meta_train_data_transformed
    # TODO: somehow compute the expected output?
    meta_features.metafeatures["LandmarkLDA"](
        X_transformed, y, logging.getLogger("Meta")
    )


def test_landmark_naive_bayes(meta_train_data_transformed):
    X_transformed, y, feat_type_transformed = meta_train_data_transformed
    # TODO: somehow compute the expected output?
    meta_features.metafeatures["LandmarkNaiveBayes"](
        X_transformed, y, logging.getLogger("Meta")
    )


def test_landmark_decision_tree(meta_train_data_transformed):
    X_transformed, y, feat_type_transformed = meta_train_data_transformed
    # TODO: somehow compute the expected output?
    meta_features.metafeatures["LandmarkDecisionTree"](
        X_transformed, y, logging.getLogger("Meta")
    )


def test_decision_node(meta_train_data_transformed):
    X_transformed, y, feat_type_transformed = meta_train_data_transformed
    # TODO: somehow compute the expected output?
    meta_features.metafeatures["LandmarkDecisionNodeLearner"](
        X_transformed, y, logging.getLogger("Meta")
    )


def test_random_node(meta_train_data_transformed):
    X_transformed, y, feat_type_transformed = meta_train_data_transformed
    # TODO: somehow compute the expected output?
    meta_features.metafeatures["LandmarkRandomNodeLearner"](
        X_transformed, y, logging.getLogger("Meta")
    )


@unittest.skip("Currently not implemented!")
def test_worst_node(meta_train_data_transformed):
    X_transformed, y, feat_type_transformed = meta_train_data_transformed
    # TODO: somehow compute the expected output?
    meta_features.metafeatures["LandmarkWorstNodeLearner"](
        X_transformed, y, logging.getLogger("Meta")
    )


def test_1NN(meta_train_data_transformed):
    X_transformed, y, feat_type_transformed = meta_train_data_transformed
    # TODO: somehow compute the expected output?
    meta_features.metafeatures["Landmark1NN"](
        X_transformed, y, logging.getLogger("Meta")
    )


def test_pca(meta_train_data_transformed):
    X_transformed, y, feat_type_transformed = meta_train_data_transformed
    meta_features.helper_functions["PCA"](X_transformed, y, logging.getLogger("Meta"))


def test_pca_95percent(meta_train_data_transformed):
    X_transformed, y, feat_type_transformed = meta_train_data_transformed
    mf = meta_features.metafeatures["PCAFractionOfComponentsFor95PercentVariance"](
        X_transformed, y, logging.getLogger("Meta")
    )
    assert pytest.approx(0.2716049382716049) == mf.value


def test_pca_kurtosis_first_pc(meta_train_data_transformed):
    X_transformed, y, feat_type_transformed = meta_train_data_transformed
    mf = meta_features.metafeatures["PCAKurtosisFirstPC"](
        X_transformed, y, logging.getLogger("Meta")
    )
    assert pytest.approx(-0.702850) != mf.value


def test_pca_skewness_first_pc(meta_train_data_transformed):
    X_transformed, y, feat_type_transformed = meta_train_data_transformed
    mf = meta_features.metafeatures["PCASkewnessFirstPC"](
        X_transformed, y, logging.getLogger("Meta")
    )
    assert pytest.approx(0.051210) != mf.value


def test_class_occurences_multilabel(multilabel_train_data):
    X, y = multilabel_train_data
    mf = meta_features.helper_functions["ClassOccurences"](
        X, y, logging.getLogger("Meta")
    )
    assert mf.value == [
        {0: 16.0, 1: 84.0},
        {0: 8.0, 1: 92.0},
        {0: 68.0, 1: 32.0},
        {0: 15.0, 1: 85.0},
        {0: 28.0, 1: 72.0},
    ]


def test_class_probability_min_multilabel(multilabel_train_data):
    X, y = multilabel_train_data
    meta_features.helper_functions.set_value(
        "ClassOccurences",
        meta_features.helper_functions["ClassOccurences"](
            X, y, logging.getLogger("Meta")
        ),
    )
    mf = meta_features.metafeatures["ClassProbabilityMin"](
        X, y, logging.getLogger("Meta")
    )
    assert pytest.approx(mf.value) == (float(8) / float(100))
    assert isinstance(mf, MetaFeatureValue)


def test_class_probability_max_multilabel(multilabel_train_data):
    X, y = multilabel_train_data
    meta_features.helper_functions.set_value(
        "ClassOccurences",
        meta_features.helper_functions["ClassOccurences"](
            X, y, logging.getLogger("Meta")
        ),
    )
    mf = meta_features.metafeatures["ClassProbabilityMax"](
        X, y, logging.getLogger("Meta")
    )
    assert pytest.approx(mf.value) == (float(92) / float(100))
    assert isinstance(mf, MetaFeatureValue)


def test_class_probability_mean_multilabel(multilabel_train_data):
    X, y = multilabel_train_data
    meta_features.helper_functions.set_value(
        "ClassOccurences",
        meta_features.helper_functions["ClassOccurences"](
            X, y, logging.getLogger("Meta")
        ),
    )
    mf = meta_features.metafeatures["ClassProbabilityMean"](
        X, y, logging.getLogger("Meta")
    )
    classes = [(16, 84), (8, 92), (68, 32), (15, 85), (28, 72)]
    probas = np.mean([np.mean(np.array(cls_)) / 100 for cls_ in classes])
    assert mf.value == pytest.approx(probas)
    assert isinstance(mf, MetaFeatureValue)


def test_number_of_classes_multilabel(multilabel_train_data):
    X, y = multilabel_train_data
    mf = meta_features.metafeatures["NumberOfClasses"](X, y, logging.getLogger("Meta"))
    assert mf.value == 5
    assert isinstance(mf, MetaFeatureValue)


def test_class_probability_std_multilabel(multilabel_train_data):
    X, y = multilabel_train_data
    meta_features.helper_functions.set_value(
        "ClassOccurences",
        meta_features.helper_functions["ClassOccurences"](
            X, y, logging.getLogger("Meta")
        ),
    )
    mf = meta_features.metafeatures["ClassProbabilitySTD"](
        X, y, logging.getLogger("Meta")
    )
    classes = [(16, 84), (8, 92), (68, 32), (15, 85), (28, 72)]
    probas = np.mean([np.std(np.array(cls_) / 100.0) for cls_ in classes])
    assert pytest.approx(mf.value) == probas
    assert isinstance(mf, MetaFeatureValue)


def test_class_entropy_multilabel(multilabel_train_data):
    X, y = multilabel_train_data
    mf = meta_features.metafeatures["ClassEntropy"](X, y, logging.getLogger("Meta"))

    classes = [(16, 84), (8, 92), (68, 32), (15, 85), (28, 72)]
    entropies = []
    for cls in classes:
        cls = np.array(cls, dtype=np.float32)
        cls = cls / sum(cls)
        entropy = -np.sum([c * np.log2(c) for c in cls])
        entropies.append(entropy)

    assert pytest.approx(mf.value) == np.mean(entropies)


def test_landmark_lda_multilabel(multilabel_train_data):
    X, y = multilabel_train_data
    mf = meta_features.metafeatures["LandmarkLDA"](X, y, logging.getLogger("Meta"))
    assert np.isfinite(mf.value)


def test_landmark_naive_bayes_multilabel(multilabel_train_data):
    X, y = multilabel_train_data
    mf = meta_features.metafeatures["LandmarkNaiveBayes"](
        X, y, logging.getLogger("Meta")
    )
    assert np.isfinite(mf.value)


def test_landmark_decision_tree_multilabel(multilabel_train_data):
    X, y = multilabel_train_data
    mf = meta_features.metafeatures["LandmarkDecisionTree"](
        X, y, logging.getLogger("Meta")
    )
    assert np.isfinite(mf.value)


def test_landmark_decision_node_multilabel(multilabel_train_data):
    X, y = multilabel_train_data
    mf = meta_features.metafeatures["LandmarkDecisionNodeLearner"](
        X, y, logging.getLogger("Meta")
    )
    assert np.isfinite(mf.value)


def test_landmark_random_node_multilabel(multilabel_train_data):
    X, y = multilabel_train_data
    mf = meta_features.metafeatures["LandmarkRandomNodeLearner"](
        X, y, logging.getLogger("Meta")
    )
    assert np.isfinite(mf.value)


def test_1NN_multilabel(multilabel_train_data):
    X, y = multilabel_train_data
    mf = meta_features.metafeatures["Landmark1NN"](X, y, logging.getLogger("TestMeta"))
    assert np.isfinite(mf.value)


def test_calculate_all_metafeatures_multilabel(multilabel_train_data):
    meta_features.helper_functions.clear()
    X, y = multilabel_train_data
    feat_type = {i: "numerical" for i in range(10)}
    mf = meta_features.calculate_all_metafeatures(
        X, y, feat_type, "Generated", logger=logging.getLogger("TestMeta")
    )
    assert 52 == len(mf.metafeature_values)


def test_calculate_all_metafeatures_same_results_across_datatypes():
    """
    This test makes sure that numpy and pandas produce the same metafeatures.
    This also is an excuse to fully test anneal dataset, and make sure
    all metafeatures work in this complex dataset
    """
    X, y = fetch_openml(data_id=2, return_X_y=True, as_frame=True)
    feat_type = {
        col: "categorical" if X[col].dtype.name == "category" else "numerical"
        for col in X.columns
    }
    mf = meta_features.calculate_all_metafeatures(
        X, y, feat_type, "2", logger=logging.getLogger("Meta")
    )
    assert 52 == len(mf.metafeature_values)
    expected = {
        "PCASkewnessFirstPC": 0.41897660337677867,
        "PCAKurtosisFirstPC": -0.677692541156901,
        "PCAFractionOfComponentsFor95PercentVariance": 0.2716049382716049,
        "ClassEntropy": 1.1898338562043977,
        "SkewnessSTD": 7.540418815675546,
        "SkewnessMean": 1.47397188548894,
        "SkewnessMax": 29.916569235579203,
        "SkewnessMin": -29.916569235579203,
        "KurtosisSTD": 152.95700852863064,
        "KurtosisMean": 57.258120199020425,
        "KurtosisMax": 893.0011148272025,
        "KurtosisMin": -1.998392219134577,
        "SymbolsSum": 49,
        "SymbolsSTD": 1.3679553264445183,
        "SymbolsMean": 1.8846153846153846,
        "SymbolsMax": 7,
        "SymbolsMin": 1,
        "ClassProbabilitySTD": 0.28282850691819206,
        "ClassProbabilityMean": 0.2,
        "ClassProbabilityMax": 0.7616926503340757,
        "ClassProbabilityMin": 0.008908685968819599,
        "InverseDatasetRatio": 23.63157894736842,
        "DatasetRatio": 0.042316258351893093,
        "RatioNominalToNumerical": 5.333333333333333,
        "RatioNumericalToNominal": 0.1875,
        "NumberOfCategoricalFeatures": 32,
        "NumberOfNumericFeatures": 6,
        "NumberOfMissingValues": 22175.0,
        "NumberOfFeaturesWithMissingValues": 29.0,
        "NumberOfInstancesWithMissingValues": 898.0,
        "NumberOfFeatures": 38.0,
        "NumberOfClasses": 5.0,
        "NumberOfInstances": 898.0,
        "LogInverseDatasetRatio": 3.162583908575814,
        "LogDatasetRatio": -3.162583908575814,
        "PercentageOfMissingValues": 0.6498358926268901,
        "PercentageOfFeaturesWithMissingValues": 0.7631578947368421,
        "PercentageOfInstancesWithMissingValues": 1.0,
        "LogNumberOfFeatures": 3.6375861597263857,
        "LogNumberOfInstances": 6.8001700683022,
    }
    assert {k: mf[k].value for k in expected.keys()} == pytest.approx(expected)

    expected_landmarks = {
        "Landmark1NN": 0.9721601489757914,
        "LandmarkRandomNodeLearner": 0.7616945996275606,
        "LandmarkDecisionNodeLearner": 0.7827932960893855,
        "LandmarkDecisionTree": 0.9899875853507139,
        "LandmarkNaiveBayes": 0.9287150837988827,
        "LandmarkLDA": 0.9610242085661079,
    }
    assert {k: mf[k].value for k in expected_landmarks.keys()} == pytest.approx(
        expected_landmarks, rel=1e-5
    )

    # Then do numpy!
    X, y = fetch_openml(data_id=2, return_X_y=True, as_frame=False)
    feat_type = {i: value for i, value in enumerate(feat_type.values())}
    mf = meta_features.calculate_all_metafeatures(
        X, y, feat_type, "2", logger=logging.getLogger("Meta")
    )
    assert {k: mf[k].value for k in expected.keys()} == pytest.approx(expected)

    # The column-reorder of pandas and numpy array are different after
    # the data preprocessing. So we cannot directly compare, and landmarking is
    # sensible to column order
    expected_landmarks["LandmarkDecisionTree"] = 0.9922098075729361
    assert {k: mf[k].value for k in expected_landmarks.keys()} == pytest.approx(
        expected_landmarks, rel=1e-5
    )



import sklearn.linear_model

from autosklearn.pipeline.components.regression.ard_regression import ARDRegression

from .test_base import BaseRegressionComponentTest


class ARDRegressionComponentTest(BaseRegressionComponentTest):

    __test__ = True

    res = dict()
    res["default_boston"] = 0.7033160711079323
    res["default_boston_iterative"] = None
    res["default_boston_sparse"] = None
    res["default_boston_iterative_sparse"] = None
    res["default_diabetes"] = 0.4172008418124077
    res["default_diabetes_iterative"] = None
    res["default_diabetes_sparse"] = None
    res["default_diabetes_iterative_sparse"] = None

    sk_mod = sklearn.linear_model.ARDRegression

    module = ARDRegression



import numpy as np
import scipy.sparse

from autosklearn.pipeline.implementations.CategoryShift import CategoryShift

import unittest


class CategoryShiftTest(unittest.TestCase):
    def test_dense(self):
        X = np.random.randint(0, 255, (3, 4))
        Y = CategoryShift().fit_transform(X)
        self.assertTrue((Y == X + 3).all())

    def test_sparse(self):
        X = scipy.sparse.csc_matrix(
            ([1, 2, 0, 4], ([0, 1, 2, 1], [3, 2, 1, 0])), shape=(3, 4)
        )
        Y = CategoryShift().fit_transform(X)
        X.data += 3
        self.assertTrue((Y.todense() == X.todense()).all())
        # Check if the sparsity stays the same before and after the transformation
        self.assertEqual(X.data.shape, Y.data.shape)
        self.assertTrue((X.indices == Y.indices).all())
        self.assertTrue((X.indptr == Y.indptr).all())

    def test_negative(self):
        X = np.array([[-1, 2], [3, 4]])
        with self.assertRaises(ValueError):
            CategoryShift().fit_transform(X)

    def test_string(self):
        X = np.array([["a", "b"], ["c", "d"]])
        with self.assertRaises(ValueError):
            CategoryShift().fit_transform(X)



# -*- encoding: utf-8 -*-
from typing import List, Optional, Tuple, Union

import logging

import numpy as np
import pandas as pd
from scipy.sparse import spmatrix
from sklearn.base import BaseEstimator
from sklearn.exceptions import NotFittedError

from autosklearn.data.feature_validator import SUPPORTED_FEAT_TYPES, FeatureValidator
from autosklearn.data.target_validator import SUPPORTED_TARGET_TYPES, TargetValidator
from autosklearn.util.logging_ import get_named_client_logger


def convert_if_sparse(
    y: SUPPORTED_TARGET_TYPES,
) -> Union[np.ndarray, List, pd.DataFrame, pd.Series]:
    """If the labels `y` are sparse, it will convert it to its dense representation

    Parameters
    ----------
    y: {array-like, sparse matrix} of shape (n_samples,) or (n_samples, n_outputs)
        The labels to 'densify' if sparse

    Returns
    -------
    np.ndarray of shape (n_samples, ) or (n_samples, n_outputs)
    """
    if isinstance(y, spmatrix):
        y_ = y.toarray()

        # For sparse one dimensional data, y.toarray will return [[1], [2], [3], ...]
        # We need to flatten this before returning it
        if y_.shape[0] == 1:
            y_ = y_.flatten()
    else:
        y_ = y

    return y_


class InputValidator(BaseEstimator):
    """
    Makes sure the input data complies with Auto-sklearn requirements.
    Categorical inputs are encoded via a Label Encoder, if the input
    is a dataframe.

    This class also perform checks for data integrity and flags the user
    via informative errors.

    Attributes
    ----------
    feat_type: Optional[List[str]] = None
        In case the dataset is not a pandas DataFrame:
            +   If provided, this list indicates which columns should be treated as
                categorical it is internally transformed into a dictionary that
                indicates a mapping from column index to categorical/numerical.
            +   If not provided, by default all columns are treated as numerical

        If the input dataset is of type pandas dataframe, this argument
        must be none, as the column type will be inferred from the pandas dtypes.

    is_classification: bool
        For classification task, this flag indicates that the target data
        should be encoded

    feature_validator: FeatureValidator
        A FeatureValidator instance used to validate and encode feature columns to match
        sklearn expectations on the data

    target_validator: TargetValidator
        A TargetValidator instance used for classification to validate and encode the
        target values
    """

    def __init__(
        self,
        feat_type: Optional[List[str]] = None,
        is_classification: bool = False,
        logger_port: Optional[int] = None,
        allow_string_features: bool = True,
    ) -> None:
        self.feat_type = feat_type
        self.is_classification = is_classification
        self.logger_port = logger_port
        if self.logger_port is not None:
            self.logger = get_named_client_logger(
                name="Validation",
                port=self.logger_port,
            )
        else:
            self.logger = logging.getLogger("Validation")

        self.allow_string_features = allow_string_features
        self.feature_validator = FeatureValidator(
            feat_type=self.feat_type,
            logger=self.logger,
            allow_string_features=self.allow_string_features,
        )
        self.target_validator = TargetValidator(
            is_classification=self.is_classification, logger=self.logger
        )
        self._is_fitted = False

    def fit(
        self,
        X_train: SUPPORTED_FEAT_TYPES,
        y_train: SUPPORTED_TARGET_TYPES,
        X_test: Optional[SUPPORTED_FEAT_TYPES] = None,
        y_test: Optional[SUPPORTED_TARGET_TYPES] = None,
    ) -> BaseEstimator:
        """
        Validates and fit a categorical encoder (if needed) to the features, and
        a encoder for targets in the case of classification. Specifically:

        For features:
        Valid data types are enforced (List, np.ndarray, pd.DataFrame, pd.Series, scipy
        sparse) as well as dimensionality checks

        If the provided data is a pandas DataFrame with categorical/boolean/int columns,
        such columns will be encoded using an Ordinal Encoder

        For targets:
        * Checks for dimensionality as well as missing values are performed.
        * If performing a classification task, the data is going to be encoded

        Parameters
        ----------
        X_train: SUPPORTED_FEAT_TYPES
            A set of features that are going to be validated (type and dimensionality
            checks). If this data contains categorical columns, an encoder is going to
            be instantiated and trained with this data.

        y_train: SUPPORTED_TARGET_TYPES
            A set of targets to encoded if the task is for classification.

        X_test: Optional[SUPPORTED_FEAT_TYPES]
            A hold out set of features used for checking

        y_test: SUPPORTED_TARGET_TYPES
            A hold out set of targets used for checking. Additionally, if the current
            task is a classification task, this y_test categories are also going to be
            used to fit a pre-processing encoding (to prevent errors on unseen classes).

        Returns
        -------
        self
        """
        # Check that the data is valid
        if np.shape(X_train)[0] != np.shape(y_train)[0]:
            raise ValueError(
                "Inconsistent number of train datapoints for features and targets,"
                " {} for features and {} for targets".format(
                    np.shape(X_train)[0],
                    np.shape(y_train)[0],
                )
            )
        if X_test is not None and np.shape(X_test)[0] != np.shape(y_test)[0]:
            raise ValueError(
                "Inconsistent number of test datapoints for features and targets,"
                " {} for features and {} for targets".format(
                    np.shape(X_test)[0],
                    np.shape(y_test)[0],
                )
            )

        self.feature_validator.fit(X_train, X_test)
        self.target_validator.fit(y_train, y_test)
        self._is_fitted = True

        return self

    def transform(
        self,
        X: SUPPORTED_FEAT_TYPES,
        y: Optional[Union[List, pd.Series, pd.DataFrame, np.ndarray]] = None,
    ) -> Tuple[Union[np.ndarray, pd.DataFrame, spmatrix], Optional[np.ndarray]]:
        """
        Transform the given target or features to a numpy array

        Parameters
        ----------
            X: SUPPORTED_FEAT_TYPES
                A set of features to transform
            y: Optional[SUPPORTED_TARGET_TYPES]
                A set of targets to transform

        Return
        ------
            np.ndarray:
                The transformed features array
            np.ndarray:
                The transformed targets array
        """
        if not self._is_fitted:
            raise NotFittedError(
                "Cannot call transform on a validator that is not fitted"
            )

        X_transformed = self.feature_validator.transform(X)
        if y is not None:
            y_transformed = self.target_validator.transform(y)
            return X_transformed, y_transformed
        else:
            return X_transformed, None
